"""
The spec types in this file specify which types of objects can be added to a
Registry. They are literally types, and used in type annotations on core
abstractions.

Though this isn't currently enforced by the types, objects that are Specs
should be serializable to YAML format and are human readable and manageable.

Currently dicts are most often used where a Spec is required by type
signatures.

Specs are always mappings. By default, specs map from an identifier string to
a mapping of key-value properties of the spec; and in some specs such as
ArgumentSetSpec, those values can themselves be mappings.

For developer convenience many functions support flattened specs, which have
the spec identifier at the same level as the rest of the spec properties.
"""

import copy
import json
from typing import Mapping, Union, Any
from agentos.identifiers import ComponentIdentifier

FlatSpec = Mapping[str, str]


NestedComponentSpec = Mapping[str, Mapping[str, str]]
ComponentSpec = Union[NestedComponentSpec, FlatSpec]


class VersionedSpec:
    SEPARATOR = "=="


class ComponentSpecKeys:
    IDENTIFIER = "identifier"
    NAME = "name"
    VERSION = "version"
    CLASS_NAME = "class_name"
    FILE_PATH = "file_path"
    DEPENDENCIES = "dependencies"


# Repo is serialized to a YAML dictionary with the following (unflatted) form:
# {repo_identifier: {repo_property_key: repo_property_val}}
NestedRepoSpec = Mapping[str, Mapping[str, str]]
RepoSpec = Union[NestedRepoSpec, FlatSpec]


class RepoSpecKeys:
    IDENTIFIER = "identifier"
    TYPE = "type"
    URL = "url"
    PATH = "path"


# A arg_set is serialized as a ArgumentSetSpec, which is a YAML dictionary
# with the following structure:
# {component_name: {entry_point_name: {arg_name: arg_val}}
#
# arg_value can be any type supported by YAML, which includes:
# scalars (numeric or string), potentially nested lists or dictionaries with
# scalars as leaf values.
#
# Note that you can have a run use complex types via the dependencies
# mechanism which allows a component to depend on other components,
# which themselves can be instances of an arbitrary Python class.
# TODO: Figure out a better type than Any for the leaf type here.
#       Specifically, one that captures the required serializability.
ArgumentSetSpec = Mapping[str, Mapping[str, Mapping[str, Any]]]
ArgumentSetSpec.identifier_key = "identifier"


RunCommandSpec = Mapping


class RunCommandSpecKeys:
    IDENTIFIER = "identifier"  # for flattened RunCommandSpec
    COMPONENT_ID = "component"
    ENTRY_POINT = "entry_point"
    PARAMETER_SET = "argument_set"


RunSpec = Mapping


class RunSpecKeys:
    IDENTIFIER = "identifier"


def flatten_spec(nested_spec: Mapping) -> Mapping:
    assert (
        len(nested_spec.keys()) == 1
    ), f"Only specs w/ one key can be flattened: {nested_spec}"
    flat_spec = {}
    for identifier, inner_spec in nested_spec.items():
        assert type(identifier) == str
        flat_spec.update(copy.deepcopy(inner_spec))

        def err(attr):
            return (
                f"'{attr}' is a reserved spec key: if it exists inside a "
                "nested spec, its value must match the spec's identifier."
            )

        if ComponentSpecKeys.IDENTIFIER in flat_spec:
            flat_spec_id = flat_spec[ComponentSpecKeys.IDENTIFIER]
            assert flat_spec_id == identifier, err("identifier")
        if ComponentSpecKeys.NAME in flat_spec:
            name = ComponentIdentifier(identifier).name
            assert flat_spec[ComponentSpecKeys.NAME] == name, err("name")
        if ComponentSpecKeys.VERSION in flat_spec:
            ver = ComponentIdentifier(identifier).version
            assert flat_spec[ComponentSpecKeys.version] == ver, err("version")
        flat_spec[ComponentSpecKeys.IDENTIFIER] = identifier
        id_parts = identifier.split(VersionedSpec.SEPARATOR)
        assert 0 < len(id_parts) <= 2, f"invalid identifier {identifier}"
        flat_spec[ComponentSpecKeys.NAME] = id_parts[0]
        if len(id_parts) == 2:
            flat_spec[ComponentSpecKeys.VERSION] = id_parts[1]
        else:
            flat_spec[ComponentSpecKeys.VERSION] = None
    return flat_spec


def unflatten_spec(
    flat_spec: Mapping, preserve_inner_identifier: bool = False
) -> Mapping:
    """
    Takes a flat spec, and returns a nested spec. A nested spec is a map
    from the spec's identifier to a map of the specs other key->value
    attributes. A flat spec moves the identifier into the inner map,
    essentially flattening the outer two dictionaries into a single dictionary.

    :param flat_spec: a flat spec to unflatten.
    :param preserve_inner_identifier: if true, do not delete 'identiifer',
        'name', or 'version' keys (and their associated values) from the inner
        part of the nested spec that is returned. Else, do remove them, i.e.,
        normalize the spec.
    :return: Nested version of ``flat_spec``.
    """
    assert ComponentSpecKeys.IDENTIFIER in flat_spec
    identifier = flat_spec[ComponentSpecKeys.IDENTIFIER]
    if VersionedSpec.SEPARATOR in identifier:
        parts = identifier.split(VersionedSpec.SEPARATOR)
        assert len(parts) == 2
        for check_key in [ComponentSpecKeys.NAME, ComponentSpecKeys.VERSION]:
            assert check_key in flat_spec, (
                f"'{check_key} must be a key in a flat spec that "
                "has a versioned identifier (i.e., that has an '==' in its "
                "identifier."
            )
    dup_spec = copy.deepcopy(flat_spec)
    if not preserve_inner_identifier:
        if ComponentSpecKeys.NAME in dup_spec:
            dup_spec.pop(ComponentSpecKeys.NAME)
        if ComponentSpecKeys.VERSION in dup_spec:
            dup_spec.pop(ComponentSpecKeys.VERSION)
        if ComponentSpecKeys.IDENTIFIER in dup_spec:
            dup_spec.pop(ComponentSpecKeys.IDENTIFIER)
    return {identifier: dup_spec}


def is_flat(spec: Mapping) -> bool:
    assert len(spec) > 0  # specs must have at least one key-value pair.
    # Nested specs have exactly one outer-most key, i.e., their identifier.
    # Flat specs have more than one (the identifier and something else).
    return not len(spec) == 1


def json_encode_flat_spec_field(spec: Mapping, field_name: str) -> Mapping:
    assert is_flat(spec)
    assert field_name in spec, f"no key '{field_name}' in spec '{spec}'."
    spec_copy = copy.deepcopy(spec)
    spec_copy[field_name] = json.encoder.JSONEncoder().encode(
        spec_copy[field_name]
    )
    return spec_copy



import tempfile
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME
from pathlib import Path
from typing import Any, Optional, Mapping
from agentos.exceptions import PythonComponentSystemException
from agentos.identifiers import RunIdentifier
from agentos.registry import Registry
from agentos.run import Run
from agentos.run_command import RunCommand
from agentos.specs import RunSpec, flatten_spec, unflatten_spec


def active_component_run(
    caller: Any, fail_if_none: bool = False
) -> Optional[Run]:
    """
    A helper function, returns the currently active ComponentRun, if it exists,
    else None. More specifically, if the caller is an object that is managed by
    a Component (i.e. if it has a __component__ attribute) that itself has an
    active_run, return that Run.

    :param caller: the managed object to fetch the active component run for.
    :param fail_if_none: if no active component run found, throw an exception
        instead of returning None.
    :return: the active component run if it exists, else None.
    """
    from agentos.component import Component

    if isinstance(caller, Component):
        component = caller
    else:
        try:
            component = caller.__component__
        except AttributeError:
            raise PythonComponentSystemException(
                "active_run() was called on an object that is not "
                "managed by a Component. Specifically, the object passed "
                "to active_run() must have a ``__component__`` attribute."
            )
    if not component.active_run:
        if fail_if_none:
            raise PythonComponentSystemException(
                "active_run() was passed an object managed by a Component "
                "with no active_run, and fail_if_no_active_run flag was "
                "True."
            )
        else:
            return None
    else:
        return component.active_run


class ComponentRun(Run):
    IS_FROZEN_KEY = "agentos.spec_is_frozen"
    IS_COMPONENT_RUN_TAG = "pcs.is_component_run"
    RUN_COMMAND_ID_KEY = "agentos.run_command_id"
    RUN_COMMAND_REGISTRY_FILENAME = "agentos.run_command_registry.yaml"
    """
    A ComponentRun represents the execution of a specific entry point of a
    specific Component with a specific ArgumentSet.
    """

    def __init__(
        self,
        run_command: RunCommand = None,
        experiment_id: str = None,
        existing_run_id: str = None,
    ) -> None:
        assert (
            run_command or existing_run_id
        ), "One of 'run_command' or 'existing_run_id' must be provided."
        super().__init__(
            experiment_id=experiment_id, existing_run_id=existing_run_id
        )
        assert not (
            run_command and existing_run_id
        ), "`run_command` cannot be passed with `existing_run_id`."
        self._run_command = None
        if run_command:
            self.set_and_log_run_command(run_command)
        else:
            self._run_command = self._fetch_run_command()
        self.set_tag(self.IS_COMPONENT_RUN_TAG, "True")
        self.set_tag(
            MLFLOW_RUN_NAME,
            f"PCS Component '{self.run_command.component.identifier.full}' "
            f"at Entry Point '{self.run_command.entry_point}'",
        )
        self._return_value = None

    @property
    def run_command(self) -> "RunCommand":
        return self._run_command

    @property
    def return_value(self) -> str:
        return self._return_value

    @property
    def is_reproducible(self) -> bool:
        return bool(self.run_command)

    @classmethod
    def from_run_command(
        cls, run_command: RunCommand, experiment_id: str = None
    ) -> "ComponentRun":
        return cls(run_command=run_command, experiment_id=experiment_id)

    def to_registry(
        self,
        registry: Registry = None,
        recurse: bool = True,
        force: bool = False,
        include_artifacts: bool = False,
    ) -> Registry:
        spec = registry.get_run_spec(self.identifier, error_if_not_found=False)
        if spec and not force:
            assert spec == self.to_spec(), (
                f"A component run spec with identifier '{self.identifier}' "
                f"already exists in registry '{registry}' and differs from "
                "the one being added. Use force=True to overwrite the "
                "existing one."
            )
        if recurse:
            self.run_command.to_registry(
                registry, recurse=recurse, force=force
            )
        super().to_registry(
            registry, include_artifacts=include_artifacts, force=force
        )

    @classmethod
    def from_registry(
        cls, registry: Registry, identifier: RunIdentifier
    ) -> "ComponentRun":
        spec = registry.get_run_spec(identifier)
        return cls.from_spec(spec, registry)

    @classmethod
    def from_spec(cls, spec: Mapping, registry: Registry) -> "ComponentRun":
        flat_spec = flatten_spec(spec)
        run_command = RunCommand.from_registry(
            registry, flat_spec["run_command"]
        )
        w_run_cmd_from_mlflow = cls(existing_run_id=flat_spec["identifier"])
        assert w_run_cmd_from_mlflow.run_command == run_command, (
            "The RunCommand object created from the MLflow Run "
            f"'{flat_spec['identifier']}' is different from the one fetched"
            f"from the registry provided '{registry}'."
        )
        return w_run_cmd_from_mlflow

    def _fetch_run_command(self) -> RunCommand:
        try:
            path = self.download_artifacts(self.RUN_COMMAND_REGISTRY_FILENAME)
        except IOError as e:
            raise IOError(
                f"RunCommand registry artifact not found in Run with id "
                f"{self._mlflow_run_id}. {repr(e)}"
            )
        assert self.RUN_COMMAND_ID_KEY in self._mlflow_run.data.tags, (
            f"{self.RUN_COMMAND_ID_KEY} not found in the tags of MLflow "
            f"run with id {self._mlflow_run_id}."
        )
        run_command_id = self._mlflow_run.data.tags[self.RUN_COMMAND_ID_KEY]
        registry = Registry.from_yaml(path)
        return RunCommand.from_registry(registry, run_command_id)

    def set_and_log_run_command(self, run_command: RunCommand) -> None:
        """
        Log a Registry YAML file for the RunCommand of this run, including
        the ArgumentSet, entry_point (i.e., function name), component ID,
        as well as the root component being run and its full
        transitive dependency graph of other components as part of this Run.
        This registry file will contain the component spec and repo spec for
        each component in the root component's dependency graph. Note that a
        Run object contains a component object and thus the root component's
        full dependency graph of other components, and as such does not depend
        on a Registry to provide reproducibility. Like a Component, a Run
        (including its entry point, argument_set, root component, and the root
        component's full dependency graph) can be dumped into a Registry for
        sharing purposes, which essentially normalizes the Run's root
        component's dependency graph into flat component specs.
        """
        assert not self._run_command
        self._run_command = run_command
        self._validate_no_run_command_logged()
        self.set_tag(self.RUN_COMMAND_ID_KEY, run_command.identifier)
        run_command_dict = run_command.to_registry().to_dict()
        self.log_dict(run_command_dict, self.RUN_COMMAND_REGISTRY_FILENAME)

    def _validate_no_run_command_logged(self):
        assert self.RUN_COMMAND_ID_KEY not in self._mlflow_run.data.tags, (
            f"{self.RUN_COMMAND_ID_KEY} already found tags of MLflow run "
            f"with id {self._mlflow_run_id}. A run_command can only be logged "
            "once per a Run."
        )
        artifact_paths = [a.path for a in self.list_artifacts()]
        assert self.RUN_COMMAND_REGISTRY_FILENAME not in artifact_paths, (
            f"An artifact with name {self.RUN_COMMAND_REGISTRY_FILENAME} "
            "has already been logged to the MLflow run with id "
            f"{self._mlflow_run_id}. A run_command can only be logged "
            "once per a Run."
        )

    def log_return_value(
        self,
        ret_val: Any,
        format: str = "pickle",
    ):
        """
        Logs the return value of an entry_point run using the specified
        serialization format.

        :param ret_val: The Python object returned by this Run to be logged.
        :param format: Valid values are 'pickle, 'json', or 'yaml'.
        """
        assert (
            not self._return_value
        ), "return_value has already been logged and can only be logged once."
        self._return_value = ret_val
        tmp_dir_path = Path(tempfile.mkdtemp())
        filename_base = tmp_dir_path / (self.identifier + "-return_value")
        if format == "pickle":
            import pickle

            filename = filename_base.parent / (filename_base.name + ".pickle")
            with open(filename, "wb") as f:
                pickle.dump(ret_val, f)
        elif format == "json":
            import json

            filename = filename_base.parent / (filename_base.name + ".json")
            with open(filename, "w") as f:
                json.dump(ret_val, f)
        elif format == "yaml":
            import yaml

            filename = filename_base.parent / (filename_base.name + ".yaml")
            with open(filename, "w") as f:
                yaml.dump(ret_val, f)
        else:
            raise PythonComponentSystemException(
                f"Invalid format provided: {format}"
            )
        self.log_artifact(str(filename))
        filename.unlink()

    @property
    def is_publishable(self) -> bool:
        # use like: filtered_tags["is_publishable"] = self.is_publishable
        try:
            return self._mlflow_run.data.tags[self.IS_FROZEN_KEY] == "True"
        except KeyError:
            return False

    def to_spec(self, flatten: bool = False) -> RunSpec:
        flat_spec = super().to_spec(flatten=True)
        assert (
            self.run_command
        ), "Every ComponentRun instance must have a run_command."
        flat_spec["run_command"] = self.run_command.identifier
        return flat_spec if flatten else unflatten_spec(flat_spec)



"""AgentOS command line interface (CLI).

The CLI allows creation of a simple template agent.
"""
import os
import sys
import yaml
import click
from datetime import datetime
from pathlib import Path
from agentos import Component
from agentos import ArgumentSet
from agentos.run import Run
from agentos.registry import Registry
from agentos.virtual_env import VirtualEnv


@click.group()
@click.version_option()
def agentos_cmd():
    pass


def _validate_agent_name(ctx, param, value):
    if " " in value or ":" in value or "/" in value:
        raise click.BadParameter("name may not contain ' ', ':', or '/'.")
    return value


_arg_component_name = click.argument(
    "component_name", metavar="COMPONENT_NAME"
)

_arg_optional_entity_id = click.argument(
    "entity_id",
    metavar="ENTITY_ID",
    required=False,
)

_arg_run_id = click.argument("run_id", type=str, metavar="RUN_ID")

_arg_dir_names = click.argument("dir_names", nargs=-1, metavar="DIR_NAMES")


_option_registry_file = click.option(
    "--registry-file",
    "-r",
    type=click.Path(exists=True),
    default="./components.yaml",
    help="Path to registry file (components.yaml).",
)

_option_use_venv = click.option(
    "--use-auto-env/--use-outer-env",
    "use_venv",
    default=True,
    help=(
        "If --use-auto-venv is passed, then AgentOS will automatically "
        "create a virtual environment under which the Component DAG "
        "will be run. If --use-outer-env is passed, AgentOS will not "
        "create a new virtual environment for the Component DAG, instead "
        "running it in the existing outer Python environment."
    ),
)


_option_agent_name = click.option(
    "--agent-name",
    "-n",
    metavar="AGENT_NAME",
    default="BasicAgent",
    callback=_validate_agent_name,
    help="This is used as the name of the MLflow Project and "
    "Conda env for all *Directory Agents* being created. "
    "AGENT_NAME may not contain ' ', ':', or '/'.",
)

_option_assume_yes = click.option(
    "--assume-yes",
    "-y",
    is_flag=True,
    help="Automatically answers 'yes' to all user prompts.",
)
_option_force = click.option(
    "--force",
    "-f",
    is_flag=True,
    default=False,
    help="Force freeze even if repo is in a bad state.",
)


@agentos_cmd.command()
@_arg_dir_names
@_option_agent_name
def init(dir_names, agent_name):
    """Initialize current (or specified) directory as an AgentOS agent.

    \b
    Arguments:
        [OPTIONAL] DIR_NAMES zero or more space separated directories to
                             initialize. They will be created if they do
                             not exist.

    Creates an agent main.py file, a conda env, and an MLflow project file
    in all directories specified, or if none are specified, then create
    the files in current directory.
    """
    dirs = [Path(".")]
    if dir_names:
        dirs = [Path(d) for d in dir_names]

    CWD_STR = "current working directory"

    for d in dirs:
        d.mkdir(parents=True, exist_ok=True)
        _instantiate_template_files(d, agent_name)
        d = CWD_STR if d == Path(".") else d
        click.echo(
            f"\nFinished initializing AgentOS agent '{agent_name}' in {d}.\n"
        )
        click.echo("To run agent:")

        registry_path = ""
        if d != CWD_STR:
            registry_path = f"-r {d}/components.yaml"
        run_cmd = f"\tagentos run agent {registry_path}\n"
        click.echo(run_cmd)


@agentos_cmd.command()
@_arg_component_name
@_option_registry_file
@click.option(
    "--entry-point",
    metavar="ENTRY_POINT",
    type=str,
    default=None,
    help="A function of the component that AgentOS Runtime will call with "
    "the specified argument set.",
)
# Copied from https://github.com/mlflow/mlflow/blob/
# ... 3958cdf9664ade34ebcf5960bee215c80efae992/mlflow/cli.py#L54
@click.option(
    "--arg-set-list",
    "-A",
    metavar="NAME=VALUE",
    multiple=True,
    help="A argument for the run, of the form -A name=value. All arguments "
    "will be passed to the entry_point function using a Python kwargs-style "
    "keyword argument https://docs.python.org/3/glossary.html#term-argument",
)
@click.option(
    "--arg-set-file",
    metavar="PARAM_FILE",
    help="A YAML file containing arguments for the entry point being run. "
    "Will be passed to the entry_point function, along with individually "
    "specified args, via a Python kwargs-style keyword argument "
    "https://docs.python.org/3/glossary.html#term-argument",
)
@_option_use_venv
def run(
    component_name,
    registry_file,
    entry_point,
    arg_set_list,
    arg_set_file,
    use_venv,
):
    venv = VirtualEnv.from_registry_file(registry_file, component_name)
    venv.set_environment_handling(use_venv)
    with venv:
        cli_arg_dict = _user_args_to_dict(arg_set_list)
        component = Component.from_registry_file(registry_file, component_name)
        arg_set = ArgumentSet.from_yaml(arg_set_file)
        entry_point = entry_point or component.get_default_entry_point()
        arg_set.update(component_name, entry_point, cli_arg_dict)
        run = component.run_with_arg_set(entry_point, arg_set)
        print(f"Run {run.identifier} recorded.", end=" ")
        print("Execute the following for details:")
        print(f"\n  agentos status {run.identifier}\n")


@agentos_cmd.command()
@_arg_optional_entity_id
@_option_registry_file
@_option_use_venv
def status(entity_id, registry_file, use_venv):
    """
    ENTITY_ID can be a Component name or a Run ID.
    """
    print(f"entity_id is {entity_id}")
    if not entity_id:
        Run.print_all_status()
    elif Run.run_exists(entity_id):
        Run.from_existing_run_id(entity_id).print_status(detailed=True)
    else:  # assume entity_id is a ComponentIdentifier
        try:
            venv = VirtualEnv.from_registry_file(registry_file, entity_id)
            venv.set_environment_handling(use_venv)
            with venv:
                c = Component.from_registry_file(registry_file, entity_id)
                c.print_status_tree()
        except LookupError:
            print(f"No Run or component found with Identifier {entity_id}.")


@agentos_cmd.command()
@_arg_optional_entity_id
def publish_run(entity_id):
    Run.from_existing_run_id(run_id=entity_id).publish()


@agentos_cmd.command()
@_arg_run_id
def rerun(run_id):
    Run.from_registry(Registry.get_default(), run_id).rerun()


@agentos_cmd.command()
@_arg_component_name
@_option_registry_file
@_option_force
@_option_use_venv
def freeze(component_name, registry_file, force, use_venv):
    """
    Creates a version of ``registry_file`` for Component
    ``component_name`` where all Components in the dependency tree are
    associated with a specific git commit.  The resulting
    ``registry_file`` can be run on any machine with AgentOS installed.

    The requirements for pinning a Component spec are as follows:
        * All Components in the dependency tree must be in git repos
        * Those git repos must have GitHub as their origin
        * The current local branch and its counterpart on origin are at
          the same commit
        * There are no uncommitted changes in the local repo
    """
    venv = VirtualEnv.from_registry_file(registry_file, component_name)
    venv.set_environment_handling(use_venv)
    with venv:
        component = Component.from_registry_file(registry_file, component_name)
        frozen_reg = component.to_frozen_registry(force=force)
        print(yaml.dump(frozen_reg.to_dict()))


@agentos_cmd.command()
@_arg_component_name
@_option_registry_file
@_option_force
@_option_use_venv
def publish(
    component_name: str, registry_file: str, force: bool, use_venv: bool
):
    """
    This command pushes the spec for component ``component_name`` (and all its
    sub-Components) to the AgentOS server.  This command will fail if any
    Component in the dependency tree cannot be frozen.
    """
    venv = VirtualEnv.from_registry_file(registry_file, component_name)
    venv.set_environment_handling(use_venv)
    with venv:
        component = Component.from_registry_file(registry_file, component_name)
        frozen_spec = component.to_frozen_registry(force=force).to_spec()
        Registry.get_default().add_component_spec(frozen_spec)


@agentos_cmd.command()
@_option_assume_yes
def clear_env_cache(assume_yes):
    VirtualEnv.clear_env_cache(assume_yes=assume_yes)


# Copied from https://github.com/mlflow/mlflow/blob/3958cdf9664ade34ebcf5960bee215c80efae992/mlflow/cli.py#L188 # noqa: E501
def _user_args_to_dict(arguments, argument_type="P"):
    user_dict = {}
    for arg in arguments:
        split = arg.split("=", maxsplit=1)
        # Docker arguments such as `t` don't require a value -> set to True if specified # noqa: E501
        if len(split) == 1 and argument_type == "A":
            name = split[0]
            value = True
        elif len(split) == 2:
            name = split[0]
            value = split[1]
        else:
            print(
                "Invalid format for -%s parameter: '%s'. "
                "Use -%s name=value." % (argument_type, arg, argument_type)
            )
            sys.exit(1)
        if name in user_dict:
            print("Repeated parameter: '%s'" % name)
            sys.exit(1)
        user_dict[name] = value
    return user_dict


def _instantiate_template_files(d, agent_name):
    AOS_PATH = Path(__file__).parent
    for file_path in _INIT_FILES:
        with open(AOS_PATH / file_path, "r") as fin:
            with open(d / file_path.name, "w") as fout:
                print(file_path)
                content = fin.read()
                now = datetime.now().strftime("%b %d, %Y %H:%M:%S")
                header = (
                    "# This file was auto-generated by `agentos init` "
                    f"on {now}."
                )
                fout.write(
                    content.format(
                        agent_name=agent_name,
                        file_header=header,
                        abs_path=d.absolute(),
                        os_sep=os.sep,
                    )
                )


_AGENT_DEF_FILE = Path("./templates/agent.py")
_ENV_DEF_FILE = Path("./templates/environment.py")
_DATASET_DEF_FILE = Path("./templates/dataset.py")
_POLICY_DEF_FILE = Path("./templates/policy.py")
_AGENT_YAML_FILE = Path("./templates/components.yaml")
_REQUIREMENTS_FILE = Path("./templates/requirements.txt")
_README_FILE = Path("./templates/README.md")


_INIT_FILES = [
    _AGENT_DEF_FILE,
    _ENV_DEF_FILE,
    _POLICY_DEF_FILE,
    _DATASET_DEF_FILE,
    _AGENT_YAML_FILE,
    _REQUIREMENTS_FILE,
    _README_FILE,
]


if __name__ == "__main__":
    agentos_cmd()



"""The ``agentos`` module provides an API for building learning agents."""

from agentos.version import VERSION as __version__  # noqa: F401
from agentos.core import (
    Agent,
    Dataset,
    Environment,
    EnvironmentSpec,
    Policy,
    Runnable,
    Trainer,
)

from agentos.registry import Registry
from agentos.component import Component
from agentos.component_run import ComponentRun, active_component_run
from agentos.repo import Repo, LocalRepo, GitHubRepo
from agentos.argument_set import ArgumentSet
from agentos.run import Run
from agentos.run_command import RunCommand
from agentos.specs import (
    ComponentSpec,
    RepoSpec,
    ArgumentSetSpec,
    RunSpec,
    flatten_spec,
    unflatten_spec,
)
from agentos.agent_run import AgentRun

__all__ = [
    "Agent",
    "AgentRun",
    "Dataset",
    "Environment",
    "EnvironmentSpec",
    "Policy",
    "Runnable",
    "Trainer",
    "Registry",
    "Repo",
    "LocalRepo",
    "GitHubRepo",
    "Component",
    "ArgumentSet",
    "RunCommand",
    "Run",
    "ComponentRun",
    "active_component_run",
    "flatten_spec",
    "unflatten_spec",
    "ComponentSpec",
    "RepoSpec",
    "ArgumentSetSpec",
    "RunSpec",
]



import os
import sys
import yaml
import shutil
import hashlib
import sysconfig
import subprocess
from pathlib import Path
from contextlib import contextmanager

from agentos.registry import Registry
from agentos.utils import AOS_GLOBAL_REQS_DIR
from agentos.identifiers import ComponentIdentifier
from agentos.specs import unflatten_spec


class VirtualEnv:
    """
    This class manages a Python virtual environment. It provides methods to
    setup, enable, and disable virtual environments as well as utility methods
    such as one to clear the whole virtual environment cache.
    """

    def __init__(self, use_venv: bool = True, venv_path: Path = None):
        self.use_venv = use_venv
        self.venv_path = venv_path
        self._saved_venv_sys_path = None
        self._venv_is_active = False
        self.set_env_cache_path(AOS_GLOBAL_REQS_DIR)
        self._py_version = f"python{sysconfig.get_python_version()}"

    def __enter__(self):
        """
        Activates the virtual environment on context entry. Use as follows:

        ```
        venv = VirtualEnv()
        with venv:
            # do something
        ```
        """
        self.activate()

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Deactivates the virtual environment on context exit."""
        self.deactivate()

    @classmethod
    def from_registry_file(
        cls, yaml_file: str, name: str, version: str = None
    ) -> "VirtualEnv":
        """
        Given a path to a yaml registry file, a Component name, and
        (optionally) a Component version, this function instantiates and
        returns a virtual environment that satisfies all requirements specified
        in the requirements_path keys of all Components in the DAG.
        """
        registry = Registry.from_yaml(yaml_file)
        return cls.from_registry(registry, name, version)

    @classmethod
    def from_registry(
        cls, registry: Registry, name: str, version: str = None
    ) -> "VirtualEnv":
        """
        Given a Registry object, a Component name, and (optionally) a Component
        version, this function instantiates and returns a virtual environment
        that satisfies all requirements specified in the requirements_path keys
        of all Components in the DAG.
        """
        identifier = ComponentIdentifier(name, version)
        venv = cls()
        venv.build_venv_for_component(registry, identifier)
        return venv

    def set_environment_handling(self, use_venv: bool) -> None:
        """
        Enables or disables virtual environment management.  If ``use_venv`` is
        set to False, all public methods of this class will be no-ops.
        """
        self.use_venv = use_venv

    def set_env_cache_path(self, env_cache_path: Path) -> None:
        """
        Allows overriding of the path of the environment cache. The environment
        cache is where all the virtual environments for Components are created.
        """
        self._env_cache_path = env_cache_path

    @staticmethod
    def clear_env_cache(
        env_cache_path: Path = None, assume_yes: bool = False
    ) -> None:
        """
        Completely removes all the virtual environments that have been created
        for Components.  Pass True to ``assume_yes`` to run non-interactively.
        """
        env_cache_path = env_cache_path or AOS_GLOBAL_REQS_DIR
        answer = None
        if assume_yes:
            answer = "y"
        else:
            answer = input(
                f"This will remove everything under {env_cache_path}.  "
                "Continue? [Y/N] "
            )
        if assume_yes or answer.lower() in ["y", "yes"]:
            shutil.rmtree(env_cache_path)
            print("Cache cleared...")
            return
        print("Aborting...")

    def _save_default_env_info(self):
        self._default_sys_path = [p for p in sys.path]
        self._default_sys_prefix = sys.prefix
        self._default_sys_exec_prefix = sys.exec_prefix
        self._default_os_virtual_env = os.environ.get("VIRTUAL_ENV")
        self._default_os_path = os.environ.get("PATH")
        self._default_os_underscore = os.environ.get("_")

    def activate(self) -> None:
        """
        Activates the virtual environment currently being managed. When
        activated, an import statement (e.g. run by a Component) will execute
        within the virtual environment.
        """
        if not self.venv_path or not self.use_venv:
            print("VirtualEnv: Running in outer Python environment")
            return
        self._save_default_env_info()
        self._set_venv_sys_path()
        self._set_venv_sys_attributes()
        self._set_venv_sys_environment()
        self._venv_is_active = True
        print(f"VirtualEnv: Running in Python venv at {self.venv_path}")

    def _set_venv_sys_path(self):
        sys_path_copy = [p for p in sys.path]
        self._clear_sys_path()
        for p in sys_path_copy:
            if p.startswith(sys.base_prefix):
                sys.path.append(p)
        if sys.platform in ["win32", "win64"]:
            sys.path.insert(0, str(self.venv_path / "Scripts"))
            sys.path.append(str(self.venv_path / "Lib" / "site-packages"))
        else:
            sys.path.insert(0, str(self.venv_path / "bin"))
            versioned_lib_path = self.venv_path / "lib" / self._py_version
            sys.path.append(str(versioned_lib_path / "site-packages"))

    def _clear_sys_path(self):
        while len(sys.path) > 0:
            sys.path.pop()

    def _set_venv_sys_attributes(self):
        sys.prefix = str(self.venv_path)
        sys.exec_prefix = str(self.venv_path)

    def _set_venv_sys_environment(self):
        os.environ["VIRTUAL_ENV"] = str(self.venv_path)
        venv_bin_path = self.venv_path / "bin"
        os.environ["PATH"] = f'{str(venv_bin_path)}:{os.environ.get("PATH")}'
        os.environ["_"] = f"{str(self.venv_path)}/bin/python"

    def deactivate(self) -> None:
        """
        Deactivates the virtual environment (i.e. re-activates the default
        environment under which AgentOS was executed).
        """
        if not self._venv_is_active:
            return
        self._set_default_sys_path()
        self._set_default_sys_attributes()
        self._set_default_sys_environment()
        self._venv_is_active = False

    def _set_default_sys_path(self):
        self._saved_venv_sys_path = [p for p in sys.path]
        self._clear_sys_path()
        for p in self._default_sys_path:
            sys.path.append(p)

    def _set_default_sys_attributes(self):
        sys.prefix = self._default_sys_prefix
        sys.exec_prefix = self._default_sys_exec_prefix

    def _set_default_sys_environment(self):
        if self._default_os_virtual_env:
            os.environ["VIRTUAL_ENV"] = self._default_os_virtual_env
        if self._default_os_path:
            os.environ["PATH"] = self._default_os_path
        if self._default_os_underscore:
            os.environ["_"] = self._default_os_underscore

    def build_venv_for_component(
        self, registry: Registry, identifier: "ComponentIdentifier"
    ) -> Path:
        """
        Creates a new virtual environment based on the requirements specified
        by the Component DAG rooted by Component ``identifier``.  Every
        ``requirements_path`` specified by a Component in the DAG will be pip
        installed by AgentOS during the creation of the virtual environment.
        If no Component in the DAG specifies a ``requirements_path``, then no
        virtual environment is created and the Component DAG will be run in the
        outer Python environment.  Virtual environments are created in the
        environment cache.
        """
        if not self.use_venv:
            return None
        req_paths = self._get_requirement_file_paths(registry, identifier)
        if not req_paths:
            return None
        return self._create_virtual_env(req_paths)

    def _get_requirement_file_paths(
        self, registry: Registry, identifier: "ComponentIdentifier"
    ) -> set:
        # Prevent circular import
        from agentos.repo import Repo

        component_specs, repo_specs = registry.get_specs_transitively_by_id(
            identifier, flatten=True
        )
        repos = {
            repo_spec["identifier"]: Repo.from_spec(
                unflatten_spec(repo_spec), registry.base_dir
            )
            for repo_spec in repo_specs
        }

        req_paths = set()
        for c_spec in component_specs:
            if "requirements_path" not in c_spec:
                continue
            repo = repos[c_spec["repo"]]
            full_req_path = repo.get_local_file_path(
                c_spec["version"], c_spec["requirements_path"]
            ).absolute()
            if not full_req_path.exists():
                error_msg = (
                    f"Requirement path {full_req_path} specified by "
                    f"Component {c_spec} does not exist."
                )
                raise Exception(error_msg)
            req_paths.add(full_req_path)
        return req_paths

    def create_virtual_env(self) -> None:
        """
        Creates the directory and objects that back the virtual environment.
        """
        assert self.venv_path is not None
        assert not self.venv_path.exists(), f"{self.venv_path} exists already!"
        subprocess.run(["virtualenv", "-p", sys.executable, self.venv_path])

    def install_requirements_file(
        self, req_path: Path, pip_flags: dict = None
    ) -> None:
        """
        Installs the requirements_file pointed at by `req_path` into the
        virtual environment. ``pip_flags`` is a dictionary of command-line
        flags to path to pip during the installation, for example:

        ```
        {'-F': 'https://example.com/foo/bar'}
        ```

        results in pip being run with the following command-line flags:

        ```
        pip install .. -F https://example.com/foo/bar
        ```
        """
        pip_flags = pip_flags or {}
        python_path = self.venv_path / "bin" / "python"
        if sys.platform in ["win32", "win64"]:
            python_path = self.venv_path / "Scripts" / "python.exe"
        cmd = [
            str(python_path),
            "-m",
            "pip",
            "install",
            "-r",
            str(req_path),
        ]
        embedded_pip_flags = self._get_embedded_pip_flags(req_path)
        for flag, value in embedded_pip_flags.items():
            cmd.append(flag)
            cmd.append(value)

        for flag, value in pip_flags.items():
            cmd.append(flag)
            cmd.append(value)
        bin_path = self.venv_path / "bin"
        if sys.platform in ["win32", "win64"]:
            bin_path = self.venv_path / "Scripts"
        component_env = {
            "SYSTEMROOT": os.environ.get("SYSTEMROOT", ""),  # win32
            "VIRTUAL_ENV": str(self.venv_path),
            "PATH": f"{str(bin_path)}:{os.environ.get('PATH')}",
        }
        subprocess.run(cmd, env=component_env)

    def _create_virtual_env(self, req_paths: set):
        sorted_req_paths = sorted(p for p in req_paths)
        to_hash = hashlib.sha256()
        to_hash.update("empty".encode("utf-8"))
        for req_path in sorted_req_paths:
            with req_path.open() as file_in:
                reqs_data = file_in.read()
                to_hash.update(reqs_data.encode("utf-8"))
        hashed = to_hash.hexdigest()
        self.venv_path = self._env_cache_path / self._py_version / hashed

        if not self.venv_path.exists():
            self.create_virtual_env()

            for req_path in sorted_req_paths:
                self.install_requirements_file(req_path)

    def _get_embedded_pip_flags(self, req_path: Path) -> dict:
        """
        Sometimes virtual environments must be created with special flag passed
        to PIP.  This parses a requirement file and finds these flags. Flags
        are specified in a commented out yaml dictionary.  For example, the
        following in a requirements file:

        ```
        # agentos_pip_cmdline:
        #   linux:
        #       "-f": "https://download.pytorch.org/whl/torch_stable.html"
        ```

        will pass `-f https://download.pytorch.org/whl/torch_stable.html` to
        pip when the requirements file is installed on linux platforms.

        The format is as follows:

        ```
        # agentos_pip_cmdline:
        #   <platform>:
        #       <flag_1>: <value_1>
        #       <flag_2>: <value_2>
        ```
        """
        with req_path.open() as file_in:
            req_data = file_in.read()
        PIP_CMDLINE_SIGIL = "agentos_pip_cmdline:"
        found_flags = False
        flag_dict = {}
        flag_lines = []
        for line in req_data.split("\n"):
            if line.startswith("#") and PIP_CMDLINE_SIGIL in line:
                found_flags = True
            if found_flags and not line.startswith("#"):
                found_flags = False
                flag_lines = [fl.lstrip("#") for fl in flag_lines]
                tmp_dict = yaml.safe_load("\n".join(flag_lines))
                platform_dict = tmp_dict[PIP_CMDLINE_SIGIL[:-1]]
                for platform, flags in platform_dict.items():
                    if platform != sys.platform:
                        continue
                    flag_dict.update(flags)
                flag_lines = []
            if found_flags:
                flag_lines.append(line)
        return flag_dict


@contextmanager
def auto_revert_venv():
    """
    Use this context manager when you need to revert the Python environment to
    whatever was in place before the managed block.  Useful in tests when an
    exception might leave the environment in an unexpected state and cause
    spurious test failures.

    Usage example::


        with auto_revert_venv():
            # Do something here that may fail, leaving the env in a bad state
        # Env guaranteed to be reset to its pre-managed-block state here
    """
    venv = VirtualEnv()
    venv._save_default_env_info()
    venv._venv_is_active = True
    try:
        yield
    finally:
        venv.deactivate()



import copy
import yaml
import json
from hashlib import sha1
from typing import TypeVar, Mapping, Dict
from agentos.specs import ArgumentSetSpec

# Use Python generics (https://mypy.readthedocs.io/en/stable/generics.html)
T = TypeVar("T")


class ArgumentSet:
    """
    This object is used to encapsulate a set of arguments that are used to
    initialize a Component dependency DAG and to run methods on this DAG.
    """

    def __init__(self, arguments: Dict = None):
        if arguments:
            for component_name, fn_map in arguments.items():
                assert isinstance(component_name, str)
                assert isinstance(fn_map, Mapping)
                for fn_name, arg_map in fn_map.items():
                    assert isinstance(fn_name, str)
                    assert isinstance(arg_map, Mapping)
                    for arg_name, arg_val in arg_map.items():
                        assert isinstance(arg_name, str)
        self._arguments = arguments if arguments else {}
        # Ensure serializability.
        assert self.to_sorted_dict_str(), "arguments dict must be serializable"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return hash(self) == hash(self)
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return int(self._sha1(), 16)

    def __str__(self) -> str:
        return self.identifier

    @property
    def identifier(self) -> str:
        return self._sha1()

    @classmethod
    def from_yaml(cls, file_path) -> "ArgumentSet":
        arguments = {}
        if file_path is not None:
            with open(file_path) as file_in:
                arguments = yaml.safe_load(file_in)
        return ArgumentSet(arguments)

    @classmethod
    def from_spec(cls, arg_set_spec: ArgumentSetSpec) -> "ArgumentSet":
        assert (
            len(arg_set_spec) == 1
        ), "arg_set_spec must be a dict with a single key-value pair"
        arg_set_spec_id, inner_arg_set = None, {}
        for key, value in arg_set_spec.items():
            arg_set_spec_id = key
            inner_arg_set = value
        new_arg_set = cls(inner_arg_set)
        assert new_arg_set.identifier == arg_set_spec_id, (
            "Since arg_set_id is a hash of the arg_set contents, the "
            f"identifier of the new arg_set {new_arg_set.identifier} "
            "should match the identifier of the spec it was loaded from "
            f"{arg_set_spec_id}, but the two ids don't match."
        )
        return new_arg_set

    def update(self, component_name: str, fn_name: str, args: Dict) -> None:
        component_args = self._arguments.get(component_name, {})
        fn_args = component_args.get(fn_name, {})
        fn_args.update(args)
        component_args[fn_name] = fn_args
        self._arguments[component_name] = component_args

    def get_component_args(self, component_name: str) -> Dict:
        return self._arguments.get(component_name, {})

    def get_function_args(
        self, component_name: str, function_name: str
    ) -> Dict:
        component_args = self.get_component_args(component_name)
        fn_args = component_args.get(function_name, {})
        return fn_args if fn_args else {}

    def get_arg(
        self, component_name: str, function_name: str, arg_key: str
    ) -> Dict:
        fn_args = self.get_function_args(component_name, function_name)
        arg = fn_args.get(arg_key, {})
        return arg if arg else {}

    def to_spec(self, flatten: bool = False) -> ArgumentSetSpec:
        inner = copy.deepcopy(self._arguments)
        if flatten:
            inner.update({ArgumentSetSpec.identifier_key: self.identifier})
            return inner
        else:
            return {str(self.identifier): inner}

    def _sha1(self) -> str:
        # Not positive if this is stable across architectures.
        # See https://stackoverflow.com/q/27522626
        return sha1(self.to_sorted_dict_str().encode("utf-8")).hexdigest()

    def to_sorted_dict_str(self) -> str:
        # See https://stackoverflow.com/a/22003440
        return json.dumps(self._arguments, sort_keys=True)



import os
import sys
import abc
import logging
import uuid
from enum import Enum
from typing import TypeVar, Dict, Tuple, Union
from pathlib import Path
from dulwich import porcelain
from dulwich.repo import Repo as PorcelainRepo
from dulwich.objectspec import parse_ref
from dulwich.objectspec import parse_commit
from dulwich.errors import NotGitRepository

from agentos.exceptions import (
    BadGitStateException,
    PythonComponentSystemException,
)
from agentos.utils import AOS_GLOBAL_REPOS_DIR
from agentos.identifiers import ComponentIdentifier, RepoIdentifier
from agentos.specs import RepoSpec, NestedRepoSpec, RepoSpecKeys, flatten_spec
from agentos.registry import Registry, InMemoryRegistry

logger = logging.getLogger(__name__)

# Use Python generics (https://mypy.readthedocs.io/en/stable/generics.html)
T = TypeVar("T")


class RepoType(Enum):
    LOCAL = "local"
    GITHUB = "github"


class Repo(abc.ABC):
    """
    Base class used to encapsulate information about where a Component
    is located.
    """

    UNKNOWN_URL = "unknown_url"

    def __init__(self, identifier: str):
        self.identifier = identifier

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Repo):
            return self.to_spec() == other.to_spec()
        return self == other

    @staticmethod
    def from_spec(spec: NestedRepoSpec, base_dir: str = None) -> "Repo":
        flat_spec = flatten_spec(spec)
        if flat_spec["type"] == RepoType.LOCAL.value:
            return LocalRepo.from_spec(spec, base_dir)
        if flat_spec["type"] == RepoType.GITHUB.value:
            return GitHubRepo.from_spec(spec)
        raise PythonComponentSystemException(
            f"Unknown repo type '{flat_spec['type']} in "
            f"repo '{flat_spec['identifier']}'"
        )

    @classmethod
    def from_registry(
        cls, registry: Registry, identifier: RepoIdentifier
    ) -> "Repo":
        return cls.from_spec(registry.get_repo_spec(identifier))

    @classmethod
    def from_github(
        cls, github_account: str, repo_name: str, identifier: str = None
    ) -> "GitHubRepo":
        if not identifier:
            identifier = f"{github_account}__{repo_name}"
        url = f"https://github.com/{github_account}/{repo_name}"
        return GitHubRepo(identifier, url)

    @abc.abstractmethod
    def to_spec(self, flatten: bool = False) -> RepoSpec:
        return NotImplementedError  # type: ignore

    def to_registry(
        self,
        registry: Registry = None,
        recurse: bool = True,  # Leave for compatibility w/ future unified spec
        force: bool = False,
    ) -> Registry:
        if not registry:
            registry = InMemoryRegistry()
        repo_spec = registry.get_repo_spec(
            self.identifier, error_if_not_found=False
        )
        if repo_spec and not force:
            assert repo_spec == self.to_spec(), (
                f"A Repo with identifier '{self.identifier}' "
                f"already exists in registry '{registry}' that differs "
                f"from the one provided."
                f"New repo spec:\n{self.to_spec()}\n\n"
                f"Existing repo spec:\n{repo_spec}"
            )
        registry.add_repo_spec(self.to_spec())
        return registry

    @abc.abstractmethod
    def get_local_repo_dir(self, version: str) -> Path:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_local_file_path(self, version: str, file_path: str) -> Path:
        raise NotImplementedError()

    def get_version_from_git(
        self,
        component_identifier: ComponentIdentifier,
        file_path: str,
        force: bool = False,
    ) -> Tuple[str, str]:
        """
        Given a path to a Component, this returns a git hash and GitHub repo
        URL where the current version of the Component is publicly accessible.
        This will raise an exception if any of the following checks are true:

        1. The Component is not in a git repo.
        2. The origin of this repo is not GitHub.
        3. The current local branch and corresponding remote branch are not at
           the same commit.
        4. There are uncommitted changes locally


        If ''force'' is True, checks 2, 3, and 4 above are ignored.
        """
        full_path = self.get_local_file_path(
            component_identifier.version, file_path
        )
        assert full_path.exists(), f"Path {full_path} does not exist"
        try:
            self.porcelain_repo = PorcelainRepo.discover(full_path)
        except NotGitRepository:
            raise BadGitStateException(
                f"No git repo with Component found: {full_path}"
            )

        self._check_for_local_changes(force)
        url = self._check_for_github_url(force)
        curr_head_hash = self._check_remote_branch_status(force)
        self.porcelain_repo = None
        return url, curr_head_hash

    def _check_for_github_url(self, force: bool) -> str:
        url = self.UNKNOWN_URL
        try:
            remote, url = porcelain.get_remote_repo(self.porcelain_repo)
        except IndexError:
            error_msg = "Could not find remote repo"
            if force:
                print(f"Warning: {error_msg}")
            else:
                raise BadGitStateException(error_msg)
        if url is None or "github.com" not in url:
            error_msg = f"Remote must be on github, not {url}"
            if force:
                print(f"Warning: {error_msg}")
            else:
                raise BadGitStateException(error_msg)
        return url

    def _check_remote_branch_status(self, force: bool) -> str:
        try:
            remote, url = porcelain.get_remote_repo(self.porcelain_repo)
        except IndexError:
            error_msg = "Unable to get remote repo"
            if force:
                print(f"Warning: {error_msg}")
                return "unknown_version"
            else:
                raise BadGitStateException(error_msg)
        REMOTE_GIT_PREFIX = "refs/remotes"
        branch = porcelain.active_branch(self.porcelain_repo).decode("UTF-8")
        full_id = f"{REMOTE_GIT_PREFIX}/{remote}/{branch}".encode("UTF-8")
        refs_dict = self.porcelain_repo.refs.as_dict()
        try:
            curr_remote_hash = refs_dict[full_id].decode("UTF-8")
        except KeyError:
            curr_remote_hash = None

        curr_head_hash = self.porcelain_repo.head().decode("UTF-8")

        if curr_head_hash != curr_remote_hash:
            print(
                f"\nBranch {remote}/{branch} "
                "current commit differs from local:"
            )
            print(f"\t{remote}/{branch}: {curr_remote_hash}")
            print(f"\tlocal/{branch}: {curr_head_hash}\n")
            error_msg = (
                f"Push your changes to {remote}/{branch} before freezing"
            )
            if force:
                print(f"Warning: {error_msg}")
            else:
                raise BadGitStateException(error_msg)
        return curr_head_hash

    def _check_for_local_changes(self, force: bool) -> None:
        # Adapted from
        # https://github.com/dulwich/dulwich/blob/master/dulwich/porcelain.py#L1200
        # 1. Get status of staged
        tracked_changes = porcelain.get_tree_changes(self.porcelain_repo)
        # 2. Get status of unstaged
        index = self.porcelain_repo.open_index()
        normalizer = self.porcelain_repo.get_blob_normalizer()
        filter_callback = normalizer.checkin_normalize
        unstaged_changes = list(
            porcelain.get_unstaged_changes(
                index, self.porcelain_repo.path, filter_callback
            )
        )

        uncommitted_changes_exist = (
            len(unstaged_changes) > 0
            or len(tracked_changes["add"]) > 0
            or len(tracked_changes["delete"]) > 0
            or len(tracked_changes["modify"]) > 0
        )
        if uncommitted_changes_exist:
            print(
                f"\nUncommitted changes: {tracked_changes} or "
                f"{unstaged_changes}\n"
            )
            error_msg = "Commit all changes before freezing"
            if force:
                print(f"Warning: {error_msg}")
            else:
                raise BadGitStateException(error_msg)

    def get_prefixed_path_from_repo_root(
        self, identifier: ComponentIdentifier, file_path: str
    ) -> Path:
        """
        Finds the 'component_path' relative to the repo containing the
        Component.  For example, if ``component_path`` is::

            /foo/bar/baz/my_component.py

        and a git repo lives in::

            /foo/bar/.git/

        then this would return::

            baz/my_component.py
        """
        full_path = self.get_local_file_path(identifier.version, file_path)
        name = full_path.name
        curr_path = full_path.parent
        path_prefix = Path()
        while curr_path != Path(curr_path.root):
            try:
                PorcelainRepo(curr_path)
                return path_prefix / name
            except NotGitRepository:
                path_prefix = curr_path.name / path_prefix
                curr_path = curr_path.parent
        raise BadGitStateException(f"Unable to find repo: {full_path}")


class GitHubRepo(Repo):
    """
    A Component with an GitHubRepo can be found on GitHub.
    """

    def __init__(self, identifier: str, url: str):
        super().__init__(identifier)
        self.type = RepoType.GITHUB
        # https repo link allows for cloning without unlocking your GitHub keys
        url = url.replace("git@github.com:", "https://github.com/")
        self.url = url
        self.local_repo_path = None
        self.porcelain_repo = None

    @classmethod
    def from_spec(cls, spec: NestedRepoSpec) -> "LocalRepo":
        flat_spec = flatten_spec(spec)
        return cls(
            flat_spec[RepoSpecKeys.IDENTIFIER],
            url=flat_spec[RepoSpecKeys.URL],
        )

    def to_spec(self, flatten: bool = False) -> Dict:
        spec = {
            self.identifier: {
                RepoSpecKeys.TYPE: self.type.value,
                RepoSpecKeys.URL: self.url,
            }
        }
        return flatten_spec(spec) if flatten else spec

    def get_local_repo_dir(self, version: str) -> Path:
        local_repo_path = self._clone_repo(version)
        self._checkout_version(local_repo_path, version)
        sys.stdout.flush()
        return local_repo_path

    def get_local_file_path(self, version: str, file_path: str) -> Path:
        local_repo_path = self.get_local_repo_dir(version)
        return (local_repo_path / file_path).absolute()

    def _clone_repo(self, version: str) -> Path:
        org_name, proj_name = self.url.split("/")[-2:]
        clone_destination = (
            AOS_GLOBAL_REPOS_DIR / org_name / proj_name / version
        )
        if not clone_destination.exists():
            clone_destination.mkdir(parents=True)
            porcelain.clone(
                source=self.url, target=str(clone_destination), checkout=True
            )
        assert clone_destination.exists(), f"Unable to clone {self.url}"
        return clone_destination

    def _checkout_version(self, local_repo_path: Path, version: str) -> None:
        to_checkout = version if version else "master"
        curr_dir = os.getcwd()
        os.chdir(local_repo_path)
        repo = porcelain.open_repo(local_repo_path)
        treeish = None
        # Is version a branch name?
        try:
            treeish = parse_ref(repo, f"origin/{to_checkout}")
        except KeyError:
            pass

        # Is version a commit hash (long or short)?
        if treeish is None:
            treeish = parse_commit(repo, to_checkout).sha().hexdigest()

        porcelain.reset(repo=repo, mode="hard", treeish=treeish)
        os.chdir(curr_dir)


# TODO: Convert LocalRepo to GitRepo and make GitHubRepo a subclass of it
#       So that under the hood all local repos are seamlessly versioned
#       and publishing any component can be super seamless by easily
#       converting a GitRepo to some default GitHubRepo where anything that
#       a user wants to share can be pushed, perhaps something like
#       github.com/my_username/pcs_repo.
class LocalRepo(Repo):
    """
    A Component with a LocalRepo can be found on your local drive.
    """

    def __init__(self, identifier: str, local_dir: Union[Path, str] = None):
        super().__init__(identifier)
        if not local_dir:
            local_dir = f"{AOS_GLOBAL_REPOS_DIR}/{uuid.uuid4()}"
        self.type = RepoType.LOCAL
        self.local_dir = Path(local_dir).absolute()
        if self.local_dir.exists():
            assert self.local_dir.is_dir(), (
                f"local_dir {self.local_dir} passed to LocalRepo.__init__() "
                f"with identifier '{self.identifier}' exists but is not a dir."
            )
        else:
            self.local_dir.mkdir(parents=True, exist_ok=True)
            logger.debug(
                f"Created local_dir {self.local_dir} for "
                f"LocalRepo {self.identifier}."
            )

    @classmethod
    def from_spec(
        cls, spec: NestedRepoSpec, base_dir: str = None
    ) -> "LocalRepo":
        flat_spec = flatten_spec(spec)
        local_path = flat_spec[RepoSpecKeys.PATH]
        if base_dir and not Path(local_path).is_absolute():
            local_path = f"{base_dir}/{local_path}"
        return cls(
            flat_spec[RepoSpecKeys.IDENTIFIER],
            local_dir=local_path,
        )

    def to_spec(self, flatten: bool = False) -> RepoSpec:
        spec = {
            self.identifier: {
                RepoSpecKeys.TYPE: self.type.value,
                RepoSpecKeys.PATH: str(self.local_dir),
            }
        }
        return flatten_spec(spec) if flatten else spec

    def get_local_repo_dir(self, version: str = None) -> Path:
        assert version is None, "LocalRepos don't support versioning."
        return self.local_dir

    def get_local_file_path(self, version: str, relative_path: str) -> Path:
        if version is not None:
            print(
                "WARNING: version was passed into get_local_path() "
                "on a LocalRepo, which means it is being ignored. "
                "If this is actually a versioned repo, use GithubRepo "
                "or another versioned Repo type."
            )
        return self.local_dir / relative_path



import sys
import uuid
import logging
import importlib
from hashlib import sha1
from pathlib import Path
from dill.source import getsource as dill_getsource
from typing import Union, TypeVar, Dict, Type, Any, Sequence
from rich import print as rich_print
from rich.tree import Tree
from agentos.run_command import RunCommand
from agentos.component_run import ComponentRun
from agentos.identifiers import ComponentIdentifier
from agentos.specs import ComponentSpec, ComponentSpecKeys, unflatten_spec
from agentos.registry import (
    Registry,
    InMemoryRegistry,
)
from agentos.exceptions import RegistryException
from agentos.repo import RepoType, Repo, LocalRepo, GitHubRepo
from agentos.argument_set import ArgumentSet
from agentos.virtual_env import VirtualEnv
from agentos.utils import parse_github_web_ui_url

logger = logging.getLogger(__name__)

# Use Python generics (https://mypy.readthedocs.io/en/stable/generics.html)
T = TypeVar("T")


class Component:
    """
    A Component is a class manager. It provides a standard way for runtime and
    code implementations to communicate about arguments, entry points, and
    dependencies.
    """

    Identifier = ComponentIdentifier

    def __init__(
        self,
        managed_cls: Type[T],
        repo: Repo,
        identifier: "Component.Identifier",
        class_name: str,
        file_path: str,
        requirements_path: str = None,
        instantiate: bool = True,
        dependencies: Dict = None,
        dunder_name: str = None,
    ):
        """
        :param managed_cls: The object this Component manages.
        :param repo: Where the code for this component's managed object is.
        :param identifier: Used to identify the Component.
        :param class_name: The name of the class that is being managed.
        :param file_path: The python module file where the managed class is
            defined.
        :param requirements_path: Optional path to a pip installable file.
        :param instantiate: Optional. If True, then get_object() return an
            instance of the managed class; if False, it returns a class object.
        :param dependencies: List of other components that self depends on.
        :param dunder_name: Name used for the pointer to this Component on any
            instances of ``managed_cls`` created by this Component.
        """
        self._managed_cls = managed_cls
        self.repo = repo
        self.identifier = identifier
        self.class_name = class_name
        self.file_path = file_path
        self.requirements_path = requirements_path
        if not class_name:
            assert (
                not instantiate
            ), "instantiate can only be True if a class_name is provided"
        self.instantiate = instantiate
        self.dependencies = dependencies if dependencies else {}
        self._dunder_name = dunder_name or "__component__"
        self._requirements = []
        self.active_run = None

    @classmethod
    def from_github_registry(
        cls,
        github_url: str,
        name: str,
        version: str = None,
        use_venv: bool = True,
    ) -> "Component":
        """
        This method gets a Component from a registry file found on GitHub.  If
        the registry file contains a LocalRepo, this method automatically
        translates that LocalRepo into a GitHubRepo.  Pass ``use_venv=False``
        if you want to import and run the Component in your existing Python
        environment.

        The ``github_url`` argument can be found by navigating to the
        registry file on the GitHub web UI.  It should look like the
        following::

            https://github.com/<project>/<repo>/{blob,raw}/<branch>/<path>
        """
        project, repo, branch, repo_path = parse_github_web_ui_url(github_url)
        version = version or branch
        repo = Repo.from_github(project, repo)
        registry = Registry.from_repo(repo, repo_path, version)
        c_version = None
        if registry.has_component_by_name(name=name, version=version):
            c_version = version
        if use_venv:
            venv = VirtualEnv.from_registry(registry, name, c_version)
            venv.activate()
        return Component.from_registry(registry, name, c_version)

    @classmethod
    def from_default_registry(
        cls, name: str, version: str = None
    ) -> "Component":
        return cls.from_registry(Registry.from_default(), name, version)

    @classmethod
    def from_registry(
        cls, registry: Registry, name: str, version: str = None
    ) -> "Component":
        """
        Returns a Component Object from the provided registry, including
        its full dependency tree of other Component Objects.
        If no Registry is provided, use the default registry.
        """
        if version:
            identifier = ComponentIdentifier(name, version)
        else:
            identifier = ComponentIdentifier.from_str(name)
        component_specs, repo_specs = registry.get_specs_transitively_by_id(
            identifier, flatten=True
        )
        repos = {
            repo_spec["identifier"]: Repo.from_spec(
                unflatten_spec(repo_spec), registry.base_dir
            )
            for repo_spec in repo_specs
        }
        components = {}
        dependencies = {}
        for component_spec in component_specs:
            component_id = ComponentIdentifier(
                component_spec["name"], component_spec["version"]
            )
            from_repo_args = {
                "repo": repos[component_spec["repo"]],
                "identifier": component_id,
                "class_name": component_spec["class_name"],
                "file_path": component_spec["file_path"],
            }
            if "requirements_path" in component_spec:
                from_repo_args["requirements_path"] = component_spec[
                    "requirements_path"
                ]
            if "instantiate" in component_spec.keys():
                from_repo_args["instantiate"] = component_spec["instantiate"]
            component = cls.from_repo(**from_repo_args)
            components[component_id] = component
            dependencies[component_id] = component_spec.get("dependencies", {})

        # Wire up the dependency graph
        for c_name, component in components.items():
            for attr_name, dependency_name in dependencies[c_name].items():
                dependency = components[dependency_name]
                component.add_dependency(dependency, attribute_name=attr_name)

        try:
            return components[identifier]
        except KeyError:
            # Try name without the version
            unversioned_components = {
                ComponentIdentifier.from_str(c_id.name): c_obj
                for c_id, c_obj in components.items()
            }
            return unversioned_components[identifier]

    @classmethod
    def from_registry_file(
        cls, yaml_file: str, name: str, version: str = None
    ) -> "Component":
        registry = Registry.from_yaml(yaml_file)
        return cls.from_registry(registry, name, version)

    @classmethod
    def from_class(
        cls,
        managed_cls: Type[T],
        identifier: str = None,
        repo: Repo = None,
        dunder_name: str = None,
        instantiate: bool = True,
    ) -> "Component":
        name = identifier if identifier else managed_cls.__name__
        if (
            managed_cls.__module__ == "__main__"
        ):  # handle classes defined in REPL.
            file_contents = dill_getsource(managed_cls)
            if repo:
                assert repo.type == RepoType.LOCAL, (
                    f"Repo '{repo.identifier}' is type {repo.type}, but must "
                    f"be {RepoType.LOCAL.value}."
                )
            else:
                repo = LocalRepo(identifier)
            sha = str(int(sha1(file_contents.encode("utf-8")).hexdigest(), 16))
            src_file = repo.get_local_repo_dir() / f"{name}-{sha}.py"
            if src_file.exists():
                print(f"Re-using existing source file {src_file}.")
            else:
                with open(src_file, "x") as f:
                    f.write(file_contents)
                print(f"Wrote new source file {src_file}.")
        else:
            managed_cls_module = sys.modules[managed_cls.__module__]
            assert hasattr(managed_cls_module, managed_cls.__name__), (
                "Components can only be created from classes that are "
                "available as an attribute of their module."
            )
            src_file = Path(managed_cls_module.__file__)
            logger.debug(
                f"Handling managed_cls {managed_cls.__name__} from existing "
                f"source file {src_file}. dir(managed_cls): \n"
            )
            repo = LocalRepo(f"{name}_repo", local_dir=src_file.parent)
            logger.debug(
                f"Created LocalRepo {repo.identifier} from existing source "
                f"file {src_file}."
            )
        return cls(
            managed_cls=managed_cls,
            repo=repo,
            identifier=Component.Identifier(name),
            class_name=managed_cls.__name__,
            file_path=src_file.name,
            instantiate=instantiate,
            dunder_name=dunder_name,
        )

    @classmethod
    def from_repo(
        cls,
        repo: Repo,
        identifier: Union[str, "Component.Identifier"],
        class_name: str,
        file_path: str,
        requirements_path: str = None,
        instantiate: bool = True,
        dunder_name: str = None,
    ) -> "Component":
        # For convenience, optionally allow 'identifier' to be passed as str.
        identifier = ComponentIdentifier.from_str(str(identifier))
        full_path = repo.get_local_file_path(identifier.version, file_path)
        assert full_path.is_file(), f"{full_path} does not exist"
        sys.path.append(str(full_path.parent))
        spec = importlib.util.spec_from_file_location(
            f"AOS_MODULE_{class_name.upper()}", str(full_path)
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        managed_cls = getattr(module, class_name)
        sys.path.pop()
        return cls(
            managed_cls=managed_cls,
            repo=repo,
            identifier=identifier,
            class_name=class_name,
            file_path=file_path,
            requirements_path=requirements_path,
            instantiate=instantiate,
            dunder_name=dunder_name,
        )

    @property
    def name(self) -> str:
        return self.identifier.name

    @property
    def version(self):
        return self.identifier.version

    def get_default_entry_point(self):
        try:
            entry_point = self._managed_cls.DEFAULT_ENTRY_POINT
        except AttributeError:
            entry_point = "run"
        return entry_point

    def run(self, entry_point: str, **kwargs):
        """
        Run an entry point with provided arguments. If you need to specify
        arguments to the init function of the managed object or any of
        its dependency components, use :py:func:run_with_arg_set:.

        :param entry_point: name of function to call on manage object.
        :param kwargs: keyword-only args to pass through to managed object
            function called entry-point.
        :return: the return value of the entry point called.
        """
        arg_set = ArgumentSet({self.name: {entry_point: kwargs}})
        run = self.run_with_arg_set(
            entry_point,
            args=arg_set,
            log_return_value=True,
        )
        return run.return_value

    def run_with_arg_set(
        self,
        entry_point: str,
        args: Union[ArgumentSet, Dict] = None,
        publish_to: Registry = None,
        log_return_value: bool = True,
        return_value_log_format: str = "yaml",
    ) -> ComponentRun:
        """
        Run the specified entry point a new instance of this Component's
        managed object given the specified args, log the results
        and return the Run object.

        :param entry_point: Name of a function to be called on a new
            instance of this component's managed object.
        :param args: A :py:func:agentos.argument_set.ArgumentSet: or
            ArgumentSet-like dict containing the entry-point arguments, and/or
            arguments to be passed to the __init__() functions of this
            component's dependents during managed object initialization.
        :param publish_to: Optionally, publish the resulting Run object
            to the provided registry.
        :param log_return_value: If True, log the return value of the entry
            point being run.
        :param return_value_log_format: Specify which format to use when
            serializing the return value. Only used if ``log_return_value``
            is True.
        """
        assert not self.active_run, (
            f"Component {self.identifier} already has an active_run, so a "
            "new run is not allowed."
        )
        if args:
            if not isinstance(args, ArgumentSet):
                args = ArgumentSet(args)
        else:
            args = ArgumentSet()
        run_command = RunCommand(self, entry_point, args)
        with ComponentRun.from_run_command(run_command) as run:
            for c in self.dependency_list():
                c.active_run = run
            # Note: get_object() adds the dunder component attribute before
            # calling __init__ on the instance.
            instance = self.get_object(arg_set=args)
            res = self.call_function_with_arg_set(instance, entry_point, args)
            if log_return_value:
                run.log_return_value(res, return_value_log_format)
            for c in self.dependency_list():
                c.active_run = None
            if publish_to:
                run.to_registry(publish_to)
            return run

    def call_function_with_arg_set(
        self, instance: Any, function_name: str, arg_set: ArgumentSet
    ) -> Any:
        fn = getattr(instance, function_name)
        assert fn is not None, f"{instance} has no attr {function_name}"
        fn_args = arg_set.get_function_args(self.name, function_name)
        print(f"Calling {self.name}.{function_name}(**{fn_args})")
        result = fn(**fn_args)
        return result

    def add_dependency(
        self, component: "Component", attribute_name: str = None
    ) -> None:
        if type(component) is not type(self):
            raise Exception("add_dependency() must be passed a Component")
        if attribute_name is None:
            attribute_name = component.name
        assert attribute_name not in self.dependencies, (
            f"{self.identifier} already has a dependency with attribute "
            f"{attribute_name}. Please use a different attribute name."
        )
        self.dependencies[attribute_name] = component

    def get_object(self, arg_set: ArgumentSet = None) -> T:
        collected = {}
        arg_set = arg_set if arg_set else ArgumentSet({})
        return self._get_object(arg_set, collected)

    def _get_object(self, arg_set: ArgumentSet, collected: dict) -> T:
        if self.name in collected:
            return collected[self.name]
        if self.instantiate:
            save_init = self._managed_cls.__init__
            self._managed_cls.__init__ = lambda self: None
            obj = self._managed_cls()
        else:
            print(f"getting {self._managed_cls} w/o instantiating ")
            obj = self._managed_cls
        for dep_attr_name, dep_component in self.dependencies.items():
            print(f"Adding {dep_attr_name} to {self.name}")
            dep_obj = dep_component._get_object(
                arg_set=arg_set, collected=collected
            )
            setattr(obj, dep_attr_name, dep_obj)
        setattr(obj, self._dunder_name, self)
        if self.instantiate:
            self._managed_cls.__init__ = save_init
            self.call_function_with_arg_set(obj, "__init__", arg_set)
        collected[self.name] = obj
        return obj

    def _handle_repo_spec(self, repos):
        existing_repo = repos.get(self.repo.name)
        if existing_repo:
            if self.repo.to_dict() != existing_repo:
                self.repo.name = str(uuid.uuid4())
        repos[self.repo.name] = self.repo.to_dict()

    def _get_versioned_dependency_dag(
        self, force: bool = False
    ) -> "Component":
        repo_url, version = self.repo.get_version_from_git(
            self.identifier, self.file_path, force
        )
        old_identifier = Component.Identifier(self.identifier.full)
        new_identifier = Component.Identifier(old_identifier.name, version)
        prefixed_file_path = self.repo.get_prefixed_path_from_repo_root(
            new_identifier, self.file_path
        )
        prefixed_reqs_path = None
        if self.requirements_path:
            prefixed_reqs_path = self.repo.get_prefixed_path_from_repo_root(
                new_identifier, self.requirements_path
            )
        clone = Component(
            managed_cls=self._managed_cls,
            repo=GitHubRepo(identifier=self.repo.identifier, url=repo_url),
            identifier=new_identifier,
            class_name=self.class_name,
            file_path=prefixed_file_path,
            requirements_path=prefixed_reqs_path,
            instantiate=self.instantiate,
            dunder_name=self._dunder_name,
        )
        for attr_name, dependency in self.dependencies.items():
            frozen_dependency = dependency._get_versioned_dependency_dag(
                force=force
            )
            clone.add_dependency(frozen_dependency, attribute_name=attr_name)
        return clone

    def to_spec(self, flatten: bool = False) -> ComponentSpec:
        dependencies = {
            k: str(v.identifier) for k, v in self.dependencies.items()
        }
        component_spec_content = {
            "repo": self.repo.identifier,
            "file_path": str(self.file_path),
            "class_name": self.class_name,
            "dependencies": dependencies,
        }
        if self.requirements_path:
            component_spec_content["requirements_path"] = str(
                self.requirements_path
            )
        if flatten:
            component_spec_content.update(
                {ComponentSpecKeys.IDENTIFIER: str(self.identifier)}
            )
            return component_spec_content
        else:
            return {str(self.identifier): component_spec_content}

    def to_registry(
        self,
        registry: Registry = None,
        recurse: bool = True,
        force: bool = False,
    ) -> Registry:
        """
        Returns a registry containing specs for this component, all of its
        transitive dependents, and the repos of all of them. Throws an
        exception if any of them already exist in the Registry that are
        different unless ``force`` is set to True.

        :param registry: Optionally, add the component spec for this component
                         and each of its transitive dependencies (which are
                         themselves components) to the specified registry.
        :param recurse: If True, check that all transitive dependencies
                        exist in the registry already, and if they don't, then
                        add them. This includes dependencies on other
                        Components as well as dependencies a repo.
                        If they do, ensure that they are equal to
                        this component's dependencies (unless ``force`` is
                        specified).
        :param force: Optionally, if a component with the same identifier
                      already exists and is different than the current one,
                      attempt to overwrite the registered one with this one.
        """
        if not registry:
            registry = InMemoryRegistry()
        # Make sure this identifier does not exist in registry yet.
        existing_spec = registry.get_component_spec(
            self.identifier, error_if_not_found=False
        )
        if existing_spec and not force:
            if existing_spec != self.to_spec():
                raise RegistryException(
                    f"Component {self.identifier} already exists in registry "
                    f"{registry} and differs from the one you're trying to "
                    f"add. Specify force=True to overwrite it.\n"
                    f"existing: {existing_spec}"
                    "\nVS\n"
                    f"new: {self.to_spec()}"
                )
        # handle dependencies on other components
        if recurse:
            for c in self.dependency_list(include_root=False):
                existing_c_spec = registry.get_component_spec(
                    name=c.name,
                    version=c.version,
                    error_if_not_found=False,
                )
                if existing_c_spec and not force:
                    if existing_c_spec != c.to_spec():
                        raise RegistryException(
                            f"Trying to register a component {c.identifier} "
                            f"that already exists in a different form:\n"
                            f"{existing_c_spec}\n"
                            f"VS\n"
                            f"{c.to_spec()}\n\n"
                            f"To overwrite, specify force=true."
                        )
                # Either component dependency not in registry already or we are
                # force adding it.
                logger.debug(
                    f"recursively adding dependent component '{c.name}'"
                )
                c.to_registry(registry, recurse=recurse, force=force)

            # Either repo not in registry already or we are force adding it.
            self.repo.to_registry(registry, recurse=recurse, force=force)
        registry.add_component_spec(self.to_spec())
        return registry

    def to_frozen_registry(self, force: bool = False) -> Registry:
        versioned = self._get_versioned_dependency_dag(force)
        return versioned.to_registry()

    def dependency_list(
        self, include_root: bool = True
    ) -> Sequence["Component"]:
        """
        Return a normalized (i.e. flat) Sequence containing all transitive
        dependencies of this component and (optionally) this component.

        :param include_root: Whether to include root component in the list.
                             If True, self is first element in list returned.
        :return: a list containing all all of the transitive dependencies
                 of this component (optionally  including the root component).
        """
        component_queue = [self]
        ret_val = set()
        while component_queue:
            component = component_queue.pop()
            if include_root or component is not self:
                ret_val.add(component)
            for dependency in component.dependencies.values():
                component_queue.append(dependency)
        return list(ret_val)

    def print_status_tree(self) -> None:
        tree = self.get_status_tree()
        rich_print(tree)

    def get_status_tree(self, parent_tree: Tree = None) -> Tree:
        self_tree = Tree(f"Component: {self.identifier.full}")
        if parent_tree is not None:
            parent_tree.add(self_tree)
        for dep_attr_name, dep_component in self.dependencies.items():
            dep_component.get_status_tree(parent_tree=self_tree)
        return self_tree



import pprint
import yaml
from pathlib import Path
from typing import Dict, Optional


AOS_GLOBAL_CONFIG_DIR = Path.home() / ".agentos"
AOS_GLOBAL_CACHE_DIR = AOS_GLOBAL_CONFIG_DIR / "cache"
AOS_GLOBAL_REQS_DIR = AOS_GLOBAL_CACHE_DIR / "requirements_cache"
AOS_GLOBAL_REPOS_DIR = AOS_GLOBAL_CACHE_DIR / "repos_cache"


def parse_github_web_ui_url(
    github_url: str,
) -> (str, Optional[str], Optional[str]):
    """
    Parses a GitHub web UI URL pointing to a project root, of form::

        https://github.com/<project>/<repo>/

    or pointing to a specific file in a project, of form::

        https://github.com/<project>/<repo>/{blob,raw}/<branch>/<path>

    This will replace an SSH URL (i.e. one that starts with
    ``git@github.com:``) into a URL that starts with ``https://github.com/``.

    This returns a 4-tuple of:

    1. The GitHub project name

    2. The GitHub repo name

    3. The branch_name or commit hash contained in the URL (``None`` if URL is
       of the project root form)

    4. The path of the file contained in the suffix of the URL (``None`` if
       URL is of the project root form)
    """
    URL_STARTS_WITH = "https://github.com"
    # https repo link allows for cloning without unlocking your GitHub keys
    github_url = github_url.replace("git@github.com:", f"{URL_STARTS_WITH}/")
    error_msg = f'URL must start "{URL_STARTS_WITH}", not "{github_url}"'
    assert github_url.startswith(URL_STARTS_WITH), error_msg
    stripped_url = github_url.replace(f"{URL_STARTS_WITH}/", "", 1)
    split_url = stripped_url.split("/")
    assert len(split_url) >= 2, f" No project or repo in url: {github_url}"
    project_name = split_url[0]
    repo_name = split_url[1]

    # Check if URL is just a link to a GitHub project root
    if len(split_url) == 2:
        return project_name, repo_name, None, None

    # If split is not *just* a project root, then it should look like
    # [<project>, <repo>, {blob/raw}, <branch>, <path_1>, <path_2>, ...]
    assert len(split_url) >= 5, f"Can't find required paths in: {github_url}"
    branch_name = split_url[3]
    repo_path = "/".join(split_url[4:])
    return project_name, repo_name, branch_name, repo_path


def generate_dummy_dev_registry(
    version_string: str = "for_tests_dummy_dev_registry",
) -> Dict:
    registry = {}
    r2d2 = _handle_acme_r2d2(version_string)
    _merge_registry_dict(registry, r2d2)
    sb3 = _handle_sb3_agent(version_string)
    _merge_registry_dict(registry, sb3)
    rando = _handle_random_agent(version_string)
    _merge_registry_dict(registry, rando)
    pprint.pprint(registry)
    return registry


def _merge_registry_dict(a, b):
    for key, val in b.items():
        tmp = a.get(key, {})
        tmp.update(val)
        a[key] = tmp


def _handle_agent(path_prefix, rename_map):
    aos_root = Path(__file__).parent.parent
    agent_spec = aos_root / path_prefix / "components.yaml"
    with open(agent_spec) as file_in:
        registry = yaml.safe_load(file_in)
    renamed = {}
    for component_name, spec in registry.get("components").items():
        spec["file_path"] = str(path_prefix / Path(spec["file_path"]))
        if "requirements_path" in spec:
            spec["requirements_path"] = str(
                path_prefix / Path(spec["requirements_path"])
            )
        renamed[rename_map[component_name]] = spec
        renamed_dependencies = {}
        for attr_name, dep_name in spec.get("dependencies", {}).items():
            renamed_dependencies[attr_name] = rename_map[dep_name]
        spec["dependencies"] = renamed_dependencies
    registry["components"] = renamed
    registry["latest_refs"] = {
        v.split("==")[0]: v.split("==")[1] for v in rename_map.values()
    }
    return registry


def _handle_random_agent(version_string):
    random_path_prefix = Path("example_agents") / Path("random")
    random_rename_map = {
        "agent": f"random_agent=={version_string}",
        "environment": f"random_corridor=={version_string}",
        "policy": f"random_policy=={version_string}",
        "dataset": f"random_dataset=={version_string}",
        "trainer": f"random_trainer=={version_string}",
        "run_manager": f"run_manager=={version_string}",
    }
    return _handle_agent(random_path_prefix, random_rename_map)


def _handle_sb3_agent(version_string):
    sb3_path_prefix = Path("example_agents") / Path("sb3_agent")
    sb3_rename_map = {
        "sb3_agent": f"sb3_ppo_agent=={version_string}",
        "environment": f"sb3_cartpole=={version_string}",
        "SB3AgentRun": f"SB3AgentRun=={version_string}",
    }
    return _handle_agent(sb3_path_prefix, sb3_rename_map)


def _handle_acme_r2d2(version_string):
    r2d2_path_prefix = Path("example_agents") / Path("acme_r2d2")
    r2d2_rename_map = {
        "agent": f"acme_r2d2_agent=={version_string}",
        "dataset": f"acme_r2d2_dataset=={version_string}",
        "environment": f"acme_cartpole=={version_string}",
        "network": f"acme_r2d2_network=={version_string}",
        "TFModelSaver": f"TFModelSaver=={version_string}",
        "policy": f"acme_r2d2_policy=={version_string}",
        "AcmeRun": f"AcmeRun=={version_string}",
        "trainer": f"acme_r2d2_trainer=={version_string}",
    }
    return _handle_agent(r2d2_path_prefix, r2d2_rename_map)


if __name__ == "__main__":
    generate_dummy_dev_registry()



class AgentosException(Exception):
    pass


class PythonComponentSystemException(Exception):
    pass


class BadGitStateException(Exception):
    pass


class RegistryException(Exception):
    pass



class ComponentIdentifier:
    """
    A Component Identifier is an immutable string-like construct that encodes a
    name and, optionally, a version as a string using the '==' separator. It
    allows referring to Components both as [name] and [name]==[version] (e.g.,
    in registries or the command line).
    """

    @staticmethod
    def from_str(identifier_string: str) -> "ComponentIdentifier":
        split_identifier = str(identifier_string).split("==")
        assert (
            len(split_identifier) <= 2
        ), f"Bad identifier: '{identifier_string}'"
        if len(split_identifier) == 1:
            return ComponentIdentifier(split_identifier[0])
        else:
            return ComponentIdentifier(
                split_identifier[0], split_identifier[1]
            )

    def __init__(
        self,
        name: str,
        version: str = None,
    ):
        assert "==" not in name, (
            f"Component.Identifier ({name} is not allowed to contain '=='."
            f"You should probably use Component.Identifier.from_str() instead."
        )
        self._name = name
        self._version = version

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def full(self) -> str:
        if self._name and self._version:
            return "==".join((self._name, self._version))
        return self._name

    def __repr__(self) -> str:
        return f"<agentos.component.Component.Identifer: {self.full}>"

    def __hash__(self) -> int:
        return hash(self.full)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.full == other.full
        elif isinstance(other, str):
            return self.full == other
        else:
            return self is other

    def __str__(self):
        return self.full


# Like MLflow, we use strings as RunIdentifiers, per
# https://github.com/mlflow/mlflow/blob/v1.22.0/mlflow/entities/run_info.py#L99
RunIdentifier = str

RunCommandIdentifier = str

RepoIdentifier = str



import pprint
from functools import partial
from pathlib import Path
from typing import Sequence
from urllib.parse import urlparse
from mlflow.entities import RunStatus
from mlflow.exceptions import MlflowException
from mlflow.tracking import MlflowClient
from mlflow.tracking.context import registry as context_registry
from agentos.identifiers import RunIdentifier
from agentos.registry import Registry
from agentos.specs import RunSpec
from agentos.specs import unflatten_spec


class Run:
    """
    Conceptually, a Run represents code execution. More specifically, a Run has
    two distinct uses. First, a Run is used to document an instance of code
    execution and record output associated with it (similar to a logger).
    Second, a Run allows for reproducibility. For this a Run can optionally
    hold a RunCommand that can be used to recreate this run, i.e., to
    perform a "re-run".

    We implement a Run that records its RunCommand as a special type of Run
    called a :py.func.agentos.run_command.ComponentRun: which is a subclass
    of ``Run``.

    A Run is similar to a logger but provides a bit more structure than loggers
    traditionally do. For example, instead of just a log-level and some free
    text, which is the typical interface for a logger, a Run allows recording
    of tags, parameters, and metrics. These each have their own semantics and
    each is represented as a key-value pair. Currently, an AgentOS Run is a
    wrapper around an MLflow Run.

    An MLflow Run is a thin container that holds an RunData and RunInfo object.
    RunInfo contains the run metadata (id, user, timestamp, etc.)
    RunData contains metrics, params, and tags; each of which is a dict.

    AgentOS Run related abstractions are encoded into an MLflowRun as follows:
    - Component Registry -> MLflow artifact file
    - Entry point string -> MLflow run tag (MlflowRun.data.tags entry)
    - ArgumentSet -> MLflow artifact file

    A Run can also contain a [pointer to a] RunCommand.

    A Run can also have pointers to other runs. These pointers can have
    different semantic meanings. They could have an "inner-outer" relationship,
    i.e., an outer run is already active when it's "inner" run starts and ends.
    Alternatively, two runs could have an "output-input dependency-depender"
    relationship, in which the depender uses the some part of dependency run's
    output as input. In this case, the relationship is causal.
    """

    # TODO: Decide if pointers to other runs should be special some how.

    _mlflow_client = MlflowClient()

    DEFAULT_EXPERIMENT_ID = "0"
    PCS_RUN_TAG = "pcs.is_run"
    # Pass calls to the following functions through to this
    # Run's mlflow_client. All of these take run_id as
    # first arg, and so the pass-through logic also binds
    # self._mlflow_run_id as the first arg of the calls.
    PASS_THROUGH_FN_PREFIXES = [
        "log",
        "set_tag",
        "list_artifacts",
        "search_runs",
        "set_terminated",
        "download_artifacts",
    ]

    def __init__(
        self,
        experiment_id: str = None,
        existing_run_id: str = None,
    ) -> None:
        """
        Run initialization can either create a new underlying MLflowRun
        or be be based on an existing underlying MLflowRun.

        If Python gracefully supported overloading constructors, it would
        make this code a lot more easy to comprehend, but as it is
        __init__() handles both cases by inspecting which arguments are
        provided.

        Because of this, we recommend using class factory methods instead
        of directly using __init__(). For example: Run.from_run_command(),
        Run.from_existing_run_id().

        :param experiment_id: Optional Experiment ID.
        :param existing_run_id: Optional Run ID.
        """
        self._return_value = None
        if existing_run_id:
            assert (
                not experiment_id
            ), "`existing_run_id` cannot be passed with `experiment_id`"
            try:
                self._mlflow_client.get_run(existing_run_id)
            except MlflowException as mlflow_exception:
                print(
                    "Error: When creating an AgentOS Run using an "
                    "existing MLflow Run ID, an MLflow run with that ID must "
                    "be available at the current tracking URI, and "
                    f"run_id {existing_run_id} is not."
                )
                raise mlflow_exception
            self._mlflow_run_id = existing_run_id
        else:  # new run
            if experiment_id:
                exp_id = experiment_id
            else:
                exp_id = self.DEFAULT_EXPERIMENT_ID
            new_run = self._mlflow_client.create_run(exp_id)
            self._mlflow_run_id = new_run.info.run_id
            self.set_tag(self.PCS_RUN_TAG, "True")
            resolved_tags = context_registry.resolve_tags()
            for tag_k, tag_v in resolved_tags.items():
                self.set_tag(tag_k, tag_v)

    @classmethod
    def run_exists(cls, run_id) -> bool:
        try:
            cls._mlflow_client.get_run(run_id)
            return True
        except MlflowException:
            return False

    @classmethod
    def get_all_runs(cls):
        mlflow_runs = cls._mlflow_client.search_runs(
            experiment_ids=[cls.DEFAULT_EXPERIMENT_ID],
            order_by=["attribute.start_time DESC"],
            filter_string=f'tag.{cls.PCS_RUN_TAG} ILIKE "%"',
        )
        res = []
        for mlflow_run in mlflow_runs:
            r = Run.from_existing_run_id(mlflow_run.info.run_id)
            res.append(r)
        return res

    @classmethod
    def from_existing_run_id(cls, run_id: RunIdentifier) -> "Run":
        return cls(existing_run_id=run_id)

    @property
    def _mlflow_run(self):
        return self._mlflow_client.get_run(self._mlflow_run_id)

    @property
    def identifier(self) -> str:
        return self._mlflow_run.info.run_id

    @property
    def data(self) -> dict:
        return self._mlflow_run.data

    @property
    def info(self) -> dict:
        return self._mlflow_run.info

    def __getattr__(self, attr_name):
        prefix_matches = [
            attr_name.startswith(x) for x in self.PASS_THROUGH_FN_PREFIXES
        ]
        if any(prefix_matches):
            try:
                mlflow_client_fn = getattr(self._mlflow_client, attr_name)
                return partial(mlflow_client_fn, self._mlflow_run_id)
            except AttributeError as e:
                raise AttributeError(
                    f"No attribute '{attr_name}' could be found in either "
                    f"'{self.__class__} or the MlflowClient object it is "
                    f"wrapping. " + str(e)
                )
        else:
            raise AttributeError(
                f"type object '{self.__class__}' has no attribute "
                f"'{attr_name}'"
            )

    def _get_artifact_paths(self) -> Sequence[Path]:
        artifacts_dir = self.download_artifacts(".")
        artifact_paths = [
            Path(artifacts_dir) / a.path for a in self.list_artifacts()
        ]
        exist = [p.exists() for p in artifact_paths]
        assert all(exist), f"Missing artifact paths: {artifact_paths}, {exist}"
        return artifact_paths

    def end(
        self, status: str = RunStatus.to_string(RunStatus.FINISHED)
    ) -> None:
        """
        This is copied and adapted from MLflow's fluent api mlflow.end_run
        """
        self.set_terminated(status)

    def print_status(self, detailed: bool = False) -> None:
        if not detailed:
            filtered_tags = {
                k: v
                for k, v in self.data.tags.items()
                if not k.startswith("mlflow.")
            }
            print(f"\tRun {self.identifier}: {filtered_tags}")
        else:
            pprint.pprint(self.to_spec())

    @staticmethod
    def print_all_status() -> None:
        runs = Run.get_all_runs()
        print("\nRuns:")
        for run in runs:
            run.print_status()
        print()

    @classmethod
    def from_registry(
        cls,
        registry: Registry,
        run_id: RunIdentifier,
    ) -> "Run":
        run_spec = registry.get_run_spec(run_id)
        return cls.from_existing_run_id(run_spec["identifier"])

    def to_registry(
        self,
        registry: Registry = None,
        force: bool = False,
        include_artifacts: bool = False,
    ) -> Registry:
        if not registry:
            from agentos.registry import InMemoryRegistry

            registry = InMemoryRegistry()
        spec = registry.get_run_spec(self.identifier, error_if_not_found=False)
        if spec and not force:
            assert spec == self.to_spec(), (
                f"A run spec with identifier '{self.identifier}' already "
                f"exists in registry '{registry}' and differs from the one "
                "being added. Use force=True to overwrite the existing one."
            )

        registry.add_run_spec(self.to_spec())
        # If we are writing to a WebRegistry, have local artifacts, and
        # include_artifacts is True, try uploading the artifact files to the
        # registry.
        if (
            include_artifacts
            and hasattr(registry, "add_run_artifacts")
            and urlparse(self.info.artifact_uri).scheme == "file"
        ):
            local_artifact_path = self.get_artifacts_dir_path()
            registry.add_run_artifacts(self.identifier, local_artifact_path)
        return registry

    def to_spec(self, flatten: bool = False) -> RunSpec:
        mlflow_dict = self._mlflow_run.to_dictionary()
        assert "identifier" not in mlflow_dict
        mlflow_dict["identifier"] = self.identifier
        return unflatten_spec(mlflow_dict) if not flatten else mlflow_dict

    def __enter__(self) -> "Run":
        return self

    def __exit__(self, type, value, traceback) -> None:
        if type:
            self.set_terminated(RunStatus.to_string(RunStatus.FAILED))
            return
        else:
            self.end()



import abc
import logging
import os
import yaml
import json
import pprint
import shutil
import tarfile
import tempfile
import requests
from pathlib import Path
from typing import Dict, Sequence, Union, Optional, TYPE_CHECKING
from dotenv import load_dotenv
from agentos.identifiers import (
    ComponentIdentifier,
    RunIdentifier,
    RepoIdentifier,
    RunCommandIdentifier,
)
from agentos.specs import (
    flatten_spec,
    unflatten_spec,
    is_flat,
    json_encode_flat_spec_field,
    RepoSpec,
    ComponentSpec,
    NestedComponentSpec,
    RunSpec,
    RunCommandSpec,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from agentos.component import Component
    from agentos.run import Run
    from agentos.repo import Repo

# add USE_LOCAL_SERVER=True to .env to talk to local server
load_dotenv()

AOS_WEB_BASE_URL = "https://aos-web.herokuapp.com"
if os.getenv("USE_LOCAL_SERVER", False) == "True":
    AOS_WEB_BASE_URL = "http://localhost:8000"
AOS_WEB_API_EXTENSION = "/api/v1"

AOS_WEB_API_ROOT = f"{AOS_WEB_BASE_URL}{AOS_WEB_API_EXTENSION}"


class Registry(abc.ABC):
    def __init__(self, base_dir: str = None):
        self.base_dir = (
            base_dir if base_dir else "."
        )  # Used for file-backed Registry types.

    @staticmethod
    def from_dict(input_dict: Dict) -> "Registry":
        return InMemoryRegistry(input_dict)

    @staticmethod
    def from_yaml(file_path: str) -> "Registry":
        with open(file_path) as file_in:
            config = yaml.safe_load(file_in)
        return InMemoryRegistry(config, base_dir=str(Path(file_path).parent))

    @classmethod
    def from_repo(
        cls, repo: "Repo", file_path: str, version: str, format: str = "yaml"
    ) -> "Registry":
        """
        Read in a registry file from an repo.

        :param repo: Repo to load registry file from.
        :param file_path: Path within Repo that registry is located, relative
            to the repo root.
        :param format: Optionally specify the format of the registry file.
        :return: a new Registry object.
        """
        assert (
            format == "yaml"
        ), "YAML is the only registry file format supported currently"
        return cls.from_yaml(repo.get_local_repo_dir(version) / file_path)

    @classmethod
    def from_default(cls):
        if not hasattr(cls, "_default_registry"):
            cls._default_registry = WebRegistry(AOS_WEB_API_ROOT)
        return cls._default_registry

    @abc.abstractmethod
    def to_dict(self) -> Dict:
        raise NotImplementedError

    def to_yaml(self, filename: str) -> None:
        with open(filename, "w") as file:
            yaml.dump(self.to_dict(), file)

    @abc.abstractmethod
    def get_repo_spec(
        self,
        repo_id: RepoIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RepoSpec]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_component_specs(
        self, filter_by_name: str = None, filter_by_version: str = None
    ) -> NestedComponentSpec:
        """
        Return dictionary of component specs in this Registry, optionally
        filtered by name and/or version; or None if none are found.

        Each Component Spec is itself a dict mapping ComponentIdentifier to a
        dict of properties that define the Component.

        Optionally, filter the list to match all filter strings provided.
        Filters can be provided on name, version, or both.

        If this registry contains zero component specs that match
        the filter criteria (if any), then an empty dictionary is returned.

        If ``filter_by_name`` and ``filter_by_version`` are provided,
        then 0 or 1 components will be returned.

        :param filter_by_name: return only components with this name.
        :param filter_by_version: return only components with this version.
        :param include_id_in_contents: add ``name`` and ``version`` fields to
               innermost dict of the ComponentSpec Dict. This denormalizes the
               spec by duplicating the ``name`` and ``version`` which are
               already included via the ComponentIdentifier Dict key.

        :returns: A dictionary of components in this registry, optionally
                  filtered by name, version, or both. If no matching
                  components are found, an empty dictionary is returned.
        """
        raise NotImplementedError

    def get_component_spec(
        self,
        name: str,
        version: str = None,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[ComponentSpec]:
        """
        Returns the component spec with ``name`` and ``version``, if it exists,
        or raise an Error if it does not. A component's name and version are
        defined as its identifier's name and version.

        Registries are not allowed to contain multiple Components with the same
        identifier. The Registry abstract base class does not enforce that all
        Components have a version (version can be None) though some
        sub-classes, such as web service backed registries, may choose to
        enforce that constraint.

        When version is unspecified or None, this function assumes that a
        Component ``c`` exists where ``c.name == name`` and ``c.version is
        None``, and throws an error otherwise.

        Subclasses of Registry may choose to provide their own (more elaborate)
        semantics for "default components". E.g., since WebRegistry does not
        allow non-versioned components, it defines its own concept of a default
        component by maintaining a separate map from component name to
        a specific version of the component, and it allows that mapping to be
        updated by users.

        :param name: The name of the component to fetch.
        :param version: Optional version of the component to fetch.
        :param flatten: If True, flatten the outermost 2 layers of nested
            dicts into a single dict. In an unflattened component spec, the
            outermost dict is from identifier (which is a string in the format
            of name[==version]) Component component properties (class_name,
            repo, etc.). In a flattened Component spec, the name and version
            are included in the same dictionary as the class_name, repo,
            dependencies, etc.
        :param error_if_not_found: Set to False to return an empty dict in
            the case that a matching component is not found in this registry.

        :returns: a ComponentSpec (i.e. a dict) matching the filter criteria
            provided, else throw an error.

        """
        components = self.get_component_specs(name, version)
        if len(components) == 0:
            if error_if_not_found:
                raise LookupError(
                    f"This registry does not contain any components that "
                    f"match your filter criteria: name:'{name}', "
                    f"version:'{version}'."
                )
            else:
                return {}
        if len(components) > 1:
            versions = [
                ComponentIdentifier.from_str(c_id).version
                for c_id in components.keys()
            ]
            version_str = "\n - ".join(versions)
            raise LookupError(
                f"This registry contains more than one component with "
                f"the name {name}. Please specify one of the following "
                f"versions:\n - {version_str}"
            )
        return flatten_spec(components) if flatten else components

    def get_component_spec_by_id(
        self,
        identifier: Union[ComponentIdentifier, str],
        flatten: bool = False,
    ) -> Optional[ComponentSpec]:
        identifier = ComponentIdentifier.from_str(str(identifier))
        return self.get_component_spec(
            identifier.name, identifier.version, flatten=flatten
        )

    def get_specs_transitively_by_id(
        self,
        identifier: Union[ComponentIdentifier, str],
        flatten: bool = True,
    ) -> (Sequence[ComponentSpec], Sequence[RepoSpec]):
        identifier = ComponentIdentifier.from_str(str(identifier))
        component_identifiers = [identifier]
        repo_specs = {}
        component_specs = {}
        while component_identifiers:
            c_id = component_identifiers.pop()
            c_spec = self.get_component_spec_by_id(c_id, flatten=flatten)
            inner_spec = c_spec if flatten else c_spec[c_id]
            component_specs[c_id] = c_spec
            repo_id = inner_spec["repo"]
            repo_spec = self.get_repo_spec(repo_id, flatten=flatten)
            repo_specs[repo_id] = repo_spec
            for d_id in inner_spec.get("dependencies", {}).values():
                component_identifiers.append(
                    ComponentIdentifier.from_str(d_id)
                )
        return list(component_specs.values()), list(repo_specs.values())

    def has_component_by_id(self, identifier: ComponentIdentifier) -> bool:
        try:
            self.get_component_spec_by_id(identifier)
        except LookupError:
            return False
        return True

    def has_component_by_name(self, name: str, version: str = None) -> bool:
        identifier = ComponentIdentifier(name, version)
        return self.has_component_by_id(identifier)

    def get_run_command_spec(
        self,
        run_command_id: RunCommandIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RunCommandSpec]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_run_spec(
        self,
        run_id: RunIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RunSpec]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_registries(self) -> Sequence:
        raise NotImplementedError

    @abc.abstractmethod
    def add_repo_spec(self, repo_spec: RepoSpec) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def add_component_spec(self, component_spec: NestedComponentSpec) -> None:
        """
        Adds a component spec to this registry. *This does not add any
        Registry Objects* to this registry.

        To register a component, use the higher the level function
        Component.register(registry) or Registry.add_component().

        :param component_spec: The ``ComponentSpec`` to register.
        """
        raise NotImplementedError

    def add_component(
        self, component: "Component", recurse: bool = True, force: bool = False
    ) -> None:
        component.to_registry(self, recurse=recurse, force=force)

    @abc.abstractmethod
    def add_run_command_spec(self, run_command_spec: RunCommandSpec) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def add_run_spec(self, run_spec: RunSpec) -> None:
        raise NotImplementedError


class InMemoryRegistry(Registry):
    """
    A mutable in-memory registry.
    """

    def __init__(self, input_dict: Dict = None, base_dir: str = None):
        super().__init__(base_dir)
        self._registry = input_dict if input_dict else {}
        if "components" not in self._registry.keys():
            self._registry["components"] = {}
        if "repos" not in self._registry.keys():
            self._registry["repos"] = {}
        if "runs" not in self._registry.keys():
            self._registry["runs"] = {}
        if "run_commands" not in self._registry.keys():
            self._registry["run_commands"] = {}
        if "registries" not in self._registry.keys():
            self._registry["registries"] = []

    def get_repo_spec(
        self,
        repo_id: RepoIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RepoSpec]:
        return self._get_spec(repo_id, "repos", flatten, error_if_not_found)

    def get_component_specs(
        self, filter_by_name: str = None, filter_by_version: str = None
    ) -> NestedComponentSpec:
        if filter_by_name or filter_by_version:
            try:
                components = {}
                for k, v in self._registry["components"].items():
                    candidate_id = ComponentIdentifier.from_str(k)
                    passes_filter = True
                    if filter_by_name and candidate_id.name != filter_by_name:
                        passes_filter = False
                    if (
                        filter_by_version
                        and candidate_id.version != filter_by_version
                    ):
                        passes_filter = False
                    if passes_filter:
                        components[k] = v
                return components
            except KeyError:
                return {}
        return self._registry["components"]

    def get_run_command_spec(
        self,
        run_command_id: RunCommandIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RunCommandSpec]:
        return self._get_spec(
            run_command_id, "run_commands", flatten, error_if_not_found
        )

    def get_run_spec(
        self,
        run_id: RunIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RunSpec]:
        return self._get_spec(run_id, "runs", flatten, error_if_not_found)

    def _get_spec(
        self,
        identifier: str,
        spec_type: str,
        flatten: bool,
        error_if_not_found: bool = True,
    ) -> Optional[Dict]:
        """
        Factor out common functionality for fetching specs from the
        internal represention of them. Because Components are special,
        (i.e., their identifiers can be versioned) they are handled
        differently.
        """
        assert spec_type in ["repos", "runs", "run_commands"]
        if identifier not in self._registry[spec_type]:
            if error_if_not_found:
                raise LookupError(
                    f"{self.__class__}l spec with identifier "
                    f"'{identifier}' not found."
                )
            else:
                return None
        spec = {identifier: self._registry[spec_type][identifier]}
        return flatten_spec(spec) if flatten else spec

    def get_registries(self) -> Sequence[Registry]:
        return self._registry["registries"]

    def add_component_spec(self, component_spec: NestedComponentSpec) -> None:
        self._registry["components"].update(component_spec)

    def add_repo_spec(self, repo_spec: RepoSpec) -> None:
        self._registry["repos"].update(repo_spec)

    def add_run_spec(self, run_spec: RunSpec) -> None:
        self._registry["runs"].update(run_spec)

    def add_run_command_spec(self, run_command_spec: RunCommandSpec) -> None:
        self._registry["run_commands"].update(run_command_spec)

    def to_dict(self) -> Dict:
        return self._registry


class WebRegistry(Registry):
    """
    A web-server backed Registry.
    """

    def __init__(self, root_api_url: str, base_dir: str = None):
        self.root_api_url = root_api_url
        self.base_dir = (
            base_dir if base_dir else "."
        )  # Used for file-backed Registry types.

    @staticmethod
    def _is_response_ok(
        response: requests.Response, error_if_not_found: bool = True
    ) -> bool:
        if response.ok:
            return True
        else:
            content = response.content
            try:
                content = json.loads(response.content)
                if type(content) == list:
                    content = content[0]
            except json.decoder.JSONDecodeError:
                try:
                    content = content.decode()
                except Exception:
                    pass
            if error_if_not_found:
                raise LookupError(response.text)
            else:
                return False

    def get_repo_spec(
        self,
        repo_id: RepoIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> Optional[RepoSpec]:
        return self._request_spec_from_web_server(
            "repo", repo_id, flatten, error_if_not_found
        )

    def get_component_specs(
        self, filter_by_name: str = None, filter_by_version: str = None
    ) -> NestedComponentSpec:
        url_filter_str = ""
        if filter_by_name:
            url_filter_str += f"name={filter_by_name}"
        if filter_by_version:
            if url_filter_str:
                url_filter_str += "&"
            url_filter_str += f"version={filter_by_version}"
        if url_filter_str:
            url_filter_str = f"?{url_filter_str}"
        component_url = f"{self.root_api_url}/components{url_filter_str}"
        component_response = requests.get(component_url)
        assert component_response.status_code == 200
        json_results = json.loads(component_response.content)
        component_specs = {}
        for c_dict in json_results["results"]:
            identifier = f"{c_dict['name']}=={c_dict['version']}"
            dep_dict = {
                d["attribute_name"]: d["dependee"]
                for d in c_dict["depender_set"]
            }
            component_specs[identifier] = {
                "repo": c_dict["repo"],
                "file_path": c_dict["file_path"],
                "class_name": c_dict["class_name"],
                "instantiate": c_dict["instantiate"],
                "dependencies": dep_dict,
            }
        return component_specs

    def get_default_component(self, name: str):
        raise NotImplementedError

    def get_run_command_spec(
        self,
        run_command_id: RunCommandIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> RunCommandSpec:
        return self._request_spec_from_web_server(
            "runcommand", run_command_id, flatten, error_if_not_found
        )

    def get_run_spec(
        self,
        run_id: RunIdentifier,
        flatten: bool = False,
        error_if_not_found: bool = True,
    ) -> RunSpec:
        return self._request_spec_from_web_server(
            "run", run_id, flatten, error_if_not_found
        )

    def get_registries(self) -> Sequence:
        raise NotImplementedError

    def add_repo_spec(self, repo_spec: RepoSpec) -> None:
        self._post_spec_to_web_server("repo", repo_spec)

    def add_component_spec(self, component_spec: NestedComponentSpec) -> None:
        flat_spec = flatten_spec(component_spec)
        flat_spec = json_encode_flat_spec_field(flat_spec, "dependencies")
        self._post_spec_to_web_server("component", unflatten_spec(flat_spec))

    def add_run_command_spec(self, run_command_spec: RunCommandSpec) -> None:
        flat_spec = flatten_spec(run_command_spec)
        flat_spec = json_encode_flat_spec_field(flat_spec, "argument_set")
        self._post_spec_to_web_server("runcommand", unflatten_spec(flat_spec))

    def add_run_spec(self, run_spec: RunSpec) -> Sequence:
        flat_spec = flatten_spec(run_spec)
        flat_spec = json_encode_flat_spec_field(flat_spec, "info")
        flat_spec = json_encode_flat_spec_field(flat_spec, "data")
        if "artifact_tarball" not in flat_spec:
            flat_spec["artifact_tarball"] = None
        self._post_spec_to_web_server("run", flat_spec)

    def add_run_artifacts(
        self, run_id: int, run_artifact_paths: Sequence[str]
    ) -> Sequence:
        try:
            tmp_dir_path = Path(tempfile.mkdtemp())
            tar_gz_path = tmp_dir_path / f"run_{run_id}_artifacts.tar.gz"
            with tarfile.open(tar_gz_path, "w:gz") as tar:
                for artifact_path in run_artifact_paths:
                    tar.add(artifact_path, arcname=artifact_path.name)
            files = {"tarball": open(tar_gz_path, "rb")}
            url = f"{self.root_api_url}/runs/{run_id}/upload_artifact/"
            response = requests.post(url, files=files)
            result = json.loads(response.content)
            return result
        finally:
            shutil.rmtree(tmp_dir_path)

    def get_run(self, run_id: str) -> "Run":
        from agentos.run import Run

        return Run.from_registry(self, run_id)

    def _request_spec_from_web_server(
        self,
        spec_type: str,
        identifier: str,
        flatten: bool,
        error_if_not_found: bool = True,
    ) -> Optional[Dict]:
        assert spec_type in ["run", "runcommand", "repo", "component"]
        req_url = f"{self.root_api_url}/{spec_type}s/{identifier}/"
        response = requests.get(req_url)
        if not self._is_response_ok(response, error_if_not_found):
            return None
        flat_spec = json.loads(response.content)
        for spec_key, spec_val in [
            (k, v)
            for k, v in flat_spec.items()
            if k.endswith("_link") or not v
        ]:
            logger.debug(
                f"Dropping field '{spec_key}: {spec_val}' from spec "
                "returned by web server."
            )
            flat_spec.pop(spec_key)
        return unflatten_spec(flat_spec) if not flatten else flat_spec

    def _post_spec_to_web_server(self, spec_type: str, spec: dict) -> None:
        """
        Handles HTTP request to backing webserver to upload a spec.
        Handles both nested and flat specs.
        """
        assert spec_type in ["run", "runcommand", "repo", "component"]
        req_url = f"{self.root_api_url}/{spec_type}s/"
        data = spec if is_flat(spec) else flatten_spec(spec)
        logger.debug(f"\nabout to post {spec_type} spec http request: {data}")
        response = requests.post(req_url, data=data)
        self._is_response_ok(response)
        result = json.loads(response.content)
        logger.debug(f"\npost {spec_type} spec http response:")
        logger.debug(pprint.pformat(result))

    def to_dict(self) -> Dict:
        raise Exception("to_dict() is not supported on WebRegistry.")



import os
from importlib.machinery import SourceFileLoader
from setuptools import setup, find_packages

version = (
    SourceFileLoader("agentos.version", os.path.join("agentos", "version.py"))
    .load_module()
    .VERSION
)

setup(
    name="agentos",
    description=(
        "AgentOS is a command line interface and python developer API "
        "for building, running, and sharing flexible learning agents."
    ),
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="Andy Konwinski",
    version=version,
    packages=find_packages(),
    install_requires=[
        "click>=7.0",
        "gym==0.17.1",
        "numpy>=1.18.5",
        "dm-env>=1.5",
        "pyyaml>=5.4.1",
        "mlflow>=1.20.2",
        "dulwich==0.20.28",
        "requests>=2.21.0",
        "python-dotenv>=0.19.1",
        "rich>=10.15.2",
        "dill>=0.3.4",
    ],
    entry_points="""
        [console_scripts]
        agentos=agentos.cli:agentos_cmd
    """,
    license="Apache License 2.0",
    python_requires=">=3.5",
    keywords="reinforcement learning ai agent",
    url="https://agentos.org",
    project_urls={
        "Source Code": "https://github.com/agentos-project/agentos",
    },
)



"""Core AgentOS classes."""
import time
from collections import namedtuple
from threading import Thread
from typing import Sequence, Optional
from agentos.agent_run import AgentRun


class MemberInitializer:
    """Takes all constructor kwargs and sets them as class members.

    For example, if MyClass is a MemberInitializer:

    a = MyClass(foo='bar')
    assert a.foo == 'bar'
    """

    def __init__(self, **kwargs):
        """Sets all the kwargs as members on the class"""
        for k, v in kwargs.items():
            setattr(self, k, v)


class Agent(MemberInitializer):
    """An Agent observes and takes actions in its Environment.

    The primary methods on an Agent are:

        * Agent.advance() - Takes on action within the Environment as
                            determined by the Agent's policy.
        * Agent.rollout() - Advances the Agent through one episode within its
                            Environment allowing it to gather experience and
                            learn.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.curr_obs = None
        self._should_reset = True
        self._active_agent_run_stack: Sequence[AgentRun] = []

    @property
    def active_agent_run(self) -> AgentRun:
        if self._active_agent_run_stack:
            return self._active_agent_run_stack[-1]
        else:
            return None

    def evaluate(
        self,
        num_episodes,
        should_learn=False,
        max_transitions=None,
        backup_dst=None,
        print_stats=True,
        parent_run=None,
    ) -> None:
        """Runs an agent specified by a given [agent_file]

        :param num_episodes: number of episodes to run the agent through
        :param should_learn: boolean, if True we will call policy.improve
        :param max_transitions: If not None, max transitions performed before
            truncating an episode.
        :param backup_dst: if specified, will print backup path to stdout
        :param print_stats: if True, will print run stats to stdout
        :param parent_run: If set, then the AgentRun created by this function
            will set this as their parent. Else, it will try to set the
            currently active component run, else it won't set a parent.

        :returns: None
        """
        all_steps = []
        if should_learn:
            self.start_agent_run("learn", parent_run)
        else:
            self.start_agent_run("evaluate", parent_run)
        for _ in range(int(num_episodes)):
            steps = self.rollout(
                should_learn=should_learn, max_transitions=max_transitions
            )
            all_steps.append(steps)
        print(f"print_stats is {print_stats}")
        print_results = (
            print_stats != "False"
            and print_stats != "false"
            and print_stats != "f"
        )
        self.end_agent_run(print_results=print_results)

    def learn(
        self,
        num_episodes,
        test_every,
        test_num_episodes,
        max_transitions=None,
    ) -> None:
        """Trains an agent by calling its learn() method in a loop."""
        num_episodes = int(num_episodes)
        test_num_episodes = int(test_num_episodes)
        # TODO: fix that ugliness!
        # Handle strings from -A
        if isinstance(test_every, str):
            if test_every == "True":
                test_every = True
            else:
                test_every = False
        test_every = int(test_every)
        run_size = test_every if test_every else num_episodes
        total_episodes = 0

        while total_episodes < num_episodes:
            if test_every:
                self.evaluate(
                    num_episodes=test_num_episodes,
                    should_learn=False,
                    max_transitions=max_transitions,
                    backup_dst=None,
                    print_stats=True,
                )
            self.evaluate(
                num_episodes=run_size,
                should_learn=True,
                max_transitions=max_transitions,
                backup_dst=None,
                print_stats=True,
            )
            total_episodes += run_size

    def start_agent_run(self, run_type: str, parent: AgentRun) -> None:
        from agentos import active_component_run  # avoid circular import

        if not parent:
            parent = active_component_run(self)
            if parent:
                assert self.environment.__component__
                assert self.environment.__component__ in (
                    parent.run_command.component.dependency_list()
                ), (
                    "This agent's environment must be in the dependency "
                    "list of the active component run."
                )
        self._active_agent_run_stack.append(
            AgentRun(run_type, parent_run=parent)
        )

    def end_agent_run(self, print_results: bool = False) -> None:
        assert self._active_agent_run_stack, "No active AgentRun to end."
        run = self._active_agent_run_stack.pop()
        run.end(print_results=print_results)

    def advance(self) -> None:
        """Takes one action within the Environment as dictated by the Policy"""
        if self._should_reset:
            self.curr_obs = self.environment.reset()
            self._should_reset = False
            self.dataset.add(None, None, self.curr_obs, None, None, {})
        action = self.policy.decide(self.curr_obs)
        prev_obs = self.curr_obs
        self.curr_obs, reward, done, info = self.environment.step(action)
        self.dataset.add(prev_obs, action, self.curr_obs, reward, done, info)
        if done:
            self._should_reset = True
        return prev_obs, action, self.curr_obs, reward, done, info

    def rollout(self, should_learn, max_transitions=None) -> int:
        """Generates one episode of transitions and allows the Agent to
        learn from its experience.

        :param should_learn: if True, then Trainer.improve() will be called
                             every time the Agent advances one step through the
                             environment and the core training metrics
                             (step_count and episode_count) will be
                             updated after the rollout.
        :param max_transitions: If not None, the episode and rollout will be
                                truncated after the specified number of
                                transitions.

        :returns: Number of transitions experienced in this episode.
        """
        done = False
        step_count = 0
        reward = 0
        while not done:
            if max_transitions and step_count > max_transitions:
                break
            _, _, _, tmp_reward, done, _ = self.advance()
            reward += tmp_reward
            step_count += 1
            if should_learn:
                self.trainer.improve(self.dataset, self.policy)
        self.active_agent_run.add_episode_data(steps=step_count, reward=reward)
        if should_learn:
            self.trainer.improve(self.dataset, self.policy)
        return step_count


class Policy(MemberInitializer):
    """Pick next action based on last observation from environment.

    Policies are used by Agents to encapsulate any state or logic necessary
    to decide on a next action given the last observation from an env.
    """

    def decide(self, observation):
        """Takes an observation and valid actions and returns next action to
        take.

        :param observation: should be in the `observation_space` of the
            environments that this policy is compatible with.
        :returns: action to take, should be in `action_space` of the
            environments that this policy is compatible with.
        """
        raise NotImplementedError


class Trainer(MemberInitializer):
    """The Trainer is responsible for improving the Policy of the Agent as
    experience is collected.

    The primary method on the Trainer is the improve() method which gets
    called for every step taken within the episode.  It is up to the
    particular implementation to decide if this tempo is appropriate for
    training.
    """

    def improve(self, dataset, policy):
        """This method updates the policy based on the experience in the
        dataset"""
        pass


class Dataset(MemberInitializer):
    """The Dataset is responsible for recording the experience of the Agent so
    that it can be used later for training.

    The primary methods on Dataset are:
        * add() - Adds a transition into the Dataset.
        * next() - Pulls a set of transitions from the Dataset for learning.
    """

    def add(self, prev_obs, action, curr_obs, reward, done, info):
        """Adds a transition into the Dataset"""
        pass

    # TODO - actually implement this in Acme/SB3 agents
    def next(self, *args, **kwargs):
        """Pulls a set of transitions from the Dataset for learning"""
        raise NotImplementedError


class Environment(MemberInitializer):
    """
    An Env inspired by OpenAI's gym.Env and DM_Env
    https://github.com/openai/gym/blob/master/gym/core.py
    https://github.com/deepmind/dm_env/blob/master/docs/index.md
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.action_space = None
        self.observation_space = None
        self.reward_range = None

    def step(self, action):
        """Perform the action in the environment."""
        raise NotImplementedError

    def reset(self):
        """Resets the environment to an initial state."""
        raise NotImplementedError

    def render(self, mode):
        raise NotImplementedError

    def close(self, mode):
        pass

    def seed(self, seed):
        raise NotImplementedError

    def get_spec(self):
        raise NotImplementedError


# https://github.com/deepmind/acme/blob/master/acme/specs.py
EnvironmentSpec = namedtuple(
    "EnvironmentSpec", ["observations", "actions", "rewards", "discounts"]
)


class Runnable:
    def run(self, hz=40, max_iters=None, as_thread=False) -> Optional[Thread]:
        """
        Run an agent, optionally in a new thread.

        If as_thread is True, agent is run in a thread, and the
        thread object is returned to the caller. The caller may
        need to call join on that that thread depending on their
        use case for this agent_run.

        :param agent: The agent object you want to run
        :param hz: Rate at which to call agent's `advance` function. If None,
            call `advance` repeatedly in a tight loop (i.e., as fast as
            possible).
        :param max_iters: Maximum times to call agent's `advance` function,
            defaults to None.
        :param as_thread: Set to True to run this agent in a new thread,
            defaults to False.
        :returns: Either a running thread (if as_thread=True) or None.
        """

        def runner():
            done = False
            iter_count = 0
            while not done:
                if max_iters and iter_count >= max_iters:
                    break
                done = self.advance()
                if hz:
                    time.sleep(1 / hz)
                iter_count += 1

        if as_thread:
            t = Thread(target=runner)
            t.start()
            return t
        else:
            runner()



from hashlib import sha1
from typing import TYPE_CHECKING
from agentos.registry import Registry
from agentos.specs import RunCommandSpec, RunCommandSpecKeys, unflatten_spec
from agentos.identifiers import RunIdentifier, RunCommandIdentifier
from agentos.run import Run

# Avoids circular imports
if TYPE_CHECKING:
    from agentos.component import Component
    from agentos.argument_set import ArgumentSet


class RunCommand:
    """
    A RunCommand contains everything required to reproducibly execute a
    Component Entry Point. Unlike a Run, a RunCommand is not concerned with the
    outputs of the execution (see :py:func:agentos.Run: for more on that.)

    You can think of a RunCommand as a glorified dictionary containing the
    pointers to arguments and versions of code necessary to reproduce the
    setting up of a component (including its dependency dag) and the execution
    of one of its entry points with a specific ArgumentSet. Whereas a Run
    itself (which may contain a RunCommand) is more like a client to a backing
    store used various types of outputs of the code being executed.

    Our concept of a RunCommand is inspired by the MLflow ``Project Run``
    abstraction. In MLflow runs of Projects (which are roughly analogous to our
    Components) are intertwined with MLflow's concept of Runs for tracking
    purposes. In MLflow, a Project Run is a wrapper around an MLflow tracking
    Run.

    In MLflow, an entry point exists in the context of a Project Run. A
    project Run uses Tags on the underlying tracking run to log all sorts of
    metadata, including the entry point, per
    https://github.com/mlflow/mlflow/blob/v1.22.0/mlflow/projects/utils.py#L225
    and
    https://github.com/mlflow/mlflow/blob/v1.22.0/mlflow/utils/mlflow_tags.py
    """

    def __init__(
        self,
        component: "Component",
        entry_point: str,
        argument_set: "ArgumentSet",
    ):
        self._component = component
        self._entry_point = entry_point
        self._argument_set = argument_set

    def __repr__(self) -> str:
        return f"<agentos.run_command.RunCommand: {self}>"

    def __hash__(self) -> int:
        return int(self._sha1(), 16)

    def _sha1(self) -> str:
        # Not positive if this is stable across architectures.
        # See https://stackoverflow.com/q/27522626
        hash_str = (
            self._component.identifier.full
            + self._entry_point
            + self._argument_set.identifier
        )
        return sha1(hash_str.encode("utf-8")).hexdigest()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        else:
            return self is other

    def __str__(self):
        return str(hash(self))

    @property
    def identifier(self) -> str:
        return str(hash(self))

    @property
    def component(self):
        return self._component

    @property
    def entry_point(self):
        return self._entry_point

    @property
    def argument_set(self):
        return self._argument_set

    def new_run(self, experiment_id: str = None):
        return Run.from_run_command(self, experiment_id=experiment_id)

    @classmethod
    def from_default_registry(cls, run_id: RunIdentifier) -> "RunCommand":
        return cls.from_registry(Registry.from_default(), run_id)

    @classmethod
    def from_registry(
        cls,
        registry: Registry,
        run_command_id: RunCommandIdentifier,
    ) -> "RunCommand":
        run_cmd_spec = registry.get_run_command_spec(run_command_id)
        return cls.from_spec(run_cmd_spec, registry)

    @classmethod
    def from_spec(
        cls, run_cmd_spec: RunCommandSpec, registry: Registry
    ) -> "RunCommand":
        assert len(run_cmd_spec) == 1
        spec_identifier, inner_spec = None, None
        for key, value in run_cmd_spec.items():
            spec_identifier = key
            inner_spec = value
        component_id = inner_spec[RunCommandSpecKeys.COMPONENT_ID]
        from agentos.component import Component
        from agentos.argument_set import ArgumentSet

        component = Component.from_registry(registry, component_id)
        arg_set = ArgumentSet.from_spec(
            inner_spec[RunCommandSpecKeys.PARAMETER_SET]
        )
        new_run_cmd = cls(
            component=component,
            entry_point=inner_spec[RunCommandSpecKeys.ENTRY_POINT],
            argument_set=arg_set,
        )
        assert new_run_cmd.identifier == spec_identifier, (
            f"Identifier of new run_command {new_run_cmd.identifier} "
            "should match identifier of spec it was loaded from "
            f"{spec_identifier}, but they don't match."
        )
        return new_run_cmd

    def publish(self) -> None:
        """
        This function is like :py:func:to_registry: but it writes the
        RunCommand to the default registry, whereas :py:func:to_registry:
        writes the RunCommand either to an explicitly provided registry object,
        or to a new InMemoryRegistry.
        """
        if not self.is_publishable:
            raise Exception("RunCommand not publishable; Spec is not frozen!")
        default_registry = Registry.from_default()
        run_id = self.to_registry(default_registry)
        print(f"Published RunCommand {run_id} to {default_registry}.")

    def to_registry(
        self,
        registry: Registry = None,
        recurse: bool = True,
        force: bool = False,
    ) -> Registry:
        """
        Returns a registry (which may optionally already exist) containing a
        run spec for this run. If recurse is True, also adds the component that
        was run to the registry by calling ``.to_registry()`` on it, and
        passing the given registry arg as well as the recurse and force args
        through to that call.

        For details on those flags, see :py:func:agentos.Component.to_registry:
        """
        if not registry:
            from agentos.registry import InMemoryRegistry

            registry = InMemoryRegistry()

        spec = registry.get_run_command_spec(
            self.identifier, error_if_not_found=False
        )
        if spec and not force:
            assert spec == self.to_spec(), (
                f"A run command spec with identifier '{self.identifier}' "
                f"already exists in registry '{registry}' and differs from "
                "the one being added. Use force=True to overwrite the "
                "existing one."
            )
        if recurse:
            self._component.to_registry(registry, recurse, force)
        registry.add_run_command_spec(self.to_spec())
        return registry

    def run(self) -> "Run":
        """
        Create a new run using the same root component, entry point, and
        params as this RunCommand.

        :return: a new RunCommand object representing the rerun.
        """
        return self.component.run_with_arg_set(
            self.entry_point, self.argument_set
        )

    def to_spec(self, flatten: bool = False) -> RunCommandSpec:
        flat_spec = {
            RunCommandSpecKeys.IDENTIFIER: self.identifier,
            RunCommandSpecKeys.COMPONENT_ID: self._component.identifier.full,
            RunCommandSpecKeys.ENTRY_POINT: self._entry_point,
            RunCommandSpecKeys.PARAMETER_SET: self._argument_set.to_spec(),
        }
        return flat_spec if flatten else unflatten_spec(flat_spec)



import statistics
from typing import Optional
from collections import namedtuple
from agentos.run import Run
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_RUN_NAME
from mlflow.entities import RunStatus

_EPISODE_KEY = "episode_count"
_STEP_KEY = "step_count"

_RUN_STATS_MEMBERS = [
    _EPISODE_KEY,
    _STEP_KEY,
    "max_reward",
    "median_reward",
    "mean_reward",
    "min_reward",
    "training_episode_count",
    "training_step_count",
]

RunStats = namedtuple("RunStats", _RUN_STATS_MEMBERS)


class AgentRun(Run):
    """
    An AgentRun provides an API that agents can use to log agent related
    data/stats/tags/etc. AgentRun can be one of two flavors (which we call
    ``run_type``), 'evaluate' and 'learn'.

    The AgentRun can contain tags that reference other AgentRuns for tracking
    the training history of an agent.

    An ``AgentRun`` inherits from ``Run``, and adds functionality specific to
    runs of agents, such as runs that *evaluate* the agent's performance in an
    environment, or runs that cause the agent to *learn* in an environment.

    Like a ``Run``, an ``AgentRun`` can be used as a context manager, so that
    the developer doesn't need to remember to mark a run as finished, for
    example:

         with AgentRun('evaluate',
                       parent_run=self.__component__.active_run) as run:
              # run an episode
              run.log_episode(
                    # episode_data
                    ...
              )
    """

    IS_AGENT_RUN_TAG = "agentos.is_agent_run"
    LEARN_KEY = "learn"
    RESET_KEY = "reset"
    RESTORE_KEY = "restore"
    EVALUATE_KEY = "evaluate"
    RUN_TYPE_TAG = "run_type"
    AGENT_NAME_KEY = "agent_name"
    ENV_NAME_KEY = "environment_name"

    def __init__(
        self,
        run_type: str,
        parent_run: Optional[str] = None,
        agent_name: Optional[str] = None,
        environment_name: Optional[str] = None,
    ) -> None:
        """
        Create a new AgentRun.

        :param run_type: must be 'evaluate' or 'learn'
        :param parent_run: Optionally, specify the identifier of another Run
            that this run is a sub-run of. Setting this will result in this
            AgentRun being visually nested under the parent_run in the MLflow
            UI.
        :param agent_name: The name of the agent component being evaluated or
            trained. Defaults to "agent".
        :param environment_name: The name of the environment component being
            evaluated or trained. Defaults to "environment".
        """
        super().__init__()
        self.parent_run = parent_run
        self.set_tag(self.IS_AGENT_RUN_TAG, "True")
        self.episode_data = []
        self.run_type = run_type
        self.agent_name = agent_name or "agent"
        self.environment_name = environment_name or "environment"

        self.set_tag(
            MLFLOW_RUN_NAME,
            (
                f"AgentOS {run_type} with Agent '{self.agent_name}' "
                f"and Env '{self.environment_name}'"
            ),
        )
        if self.parent_run:
            self.set_tag(MLFLOW_PARENT_RUN_ID, self.parent_run.info.run_id)

        self.log_run_type(self.run_type)
        self.log_agent_name(self.agent_name)
        self.log_environment_name(self.environment_name)

    def log_run_type(self, run_type: str) -> None:
        self.run_type = run_type
        self.set_tag(self.RUN_TYPE_TAG, self.run_type)

    def log_agent_name(self, agent_name: str) -> None:
        self.log_param(self.AGENT_NAME_KEY, agent_name)

    def log_environment_name(self, environment_name: str) -> None:
        self.log_param(self.ENV_NAME_KEY, environment_name)

    def log_run_metrics(self):
        assert self.episode_data, "No episode data!"
        run_stats = self._get_run_stats()
        for key in _RUN_STATS_MEMBERS:
            val = getattr(run_stats, key)
            self.log_metric(key, val)

    def get_training_info(self) -> (int, int):
        runs = self.get_all_runs()
        total_episodes = 0
        total_steps = 0
        for run in runs:
            if run.data.tags.get(self.RUN_TYPE_TAG) == self.LEARN_KEY:
                total_episodes += int(run.data.metrics.get(_EPISODE_KEY, 0))
                total_steps += int(run.data.metrics.get(_STEP_KEY, 0))
        return total_episodes, total_steps

    def print_results(self):
        if not self.episode_data:
            return
        run_stats = self._get_run_stats()
        if self.run_type == self.LEARN_KEY:
            print(
                "\nTraining results over "
                f"{run_stats.episode_count} episodes:"
            )
            print(
                "\tOverall agent was trained on "
                f"{run_stats.training_step_count} transitions over "
                f"{run_stats.training_episode_count} episodes"
            )
        else:
            print(
                "\nBenchmark results over "
                f"{run_stats.episode_count} episodes:"
            )
            print(
                "\tBenchmarked agent was trained on "
                f"{run_stats.training_step_count} transitions over "
                f"{run_stats.training_episode_count} episodes"
            )
        print(
            f"\tMax reward over {run_stats.episode_count} episodes: "
            f"{run_stats.max_reward}"
        )
        print(
            f"\tMean reward over {run_stats.episode_count} episodes: "
            f"{run_stats.mean_reward}"
        )
        print(
            f"\tMedian reward over {run_stats.episode_count} episodes: "
            f"{run_stats.median_reward}"
        )
        print(
            f"\tMin reward over {run_stats.episode_count} episodes: "
            f"{run_stats.min_reward}"
        )
        print()

    def _get_run_stats(self):
        episode_lengths = [d["steps"] for d in self.episode_data]
        episode_returns = [d["reward"] for d in self.episode_data]
        training_episodes, training_steps = self.get_training_info()
        return RunStats(
            episode_count=len(self.episode_data),
            step_count=sum(episode_lengths),
            max_reward=max(episode_returns),
            mean_reward=statistics.mean(episode_returns),
            median_reward=statistics.median(episode_returns),
            min_reward=min(episode_returns),
            training_episode_count=training_episodes,
            training_step_count=training_steps,
        )

    def add_episode_data(self, steps: int, reward: float):
        self.episode_data.append(
            {
                "steps": steps,
                "reward": reward,
            }
        )

    def end(
        self,
        status: str = RunStatus.to_string(RunStatus.FINISHED),
        print_results: bool = True,
    ) -> None:
        super().end(status)
        self.log_run_metrics()
        if print_results:
            self.print_results()

    def __enter__(self) -> "AgentRun":
        return self

    def __exit__(self, type, value, traceback) -> None:
        super().__exit__(type, value, traceback)
